# coding: utf-8

# In[ ]:


import random

import matplotlib.pyplot as plt

plt.style.use('science')

# 变量筛选
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier

# from JustinScoreCardPy import model_evaluation


# xgboost筛选变量
def select_xgboost(df, target, imp_num=None):
    """
    df:数据集
    target:目标变量的字段名
    imp_num:筛选变量的个数

    return:
    xg_fea_imp:变量的特征重要性
    xg_select_col:筛选出的变量
    """
    x = df.drop([target], axis=1)
    y = df[target]
    xgmodel = XGBClassifier(random_state=0)
    xgmodel = xgmodel.fit(x, y, eval_metric='auc')
    xg_fea_imp = pd.DataFrame({'col': list(x.columns),
                               'imp': xgmodel.feature_importances_})
    xg_fea_imp = xg_fea_imp.sort_values('imp', ascending=False).reset_index(drop=True).iloc[:imp_num, :]
    xg_select_col = list(xg_fea_imp.col)
    return xg_fea_imp, xg_select_col


# 随机森林筛选变量 
def select_rf(df, target, imp_num=None):
    """
    df:数据集
    target:目标变量的字段名
    imp_num:筛选变量的个数

    return:
    rf_fea_imp:变量的特征重要性
    rf_select_col:筛选出的变量
    """
    x = df.drop([target], axis=1)
    y = df[target]
    rfmodel = RandomForestClassifier(random_state=0)
    rfmodel = rfmodel.fit(x, y)
    rf_fea_imp = pd.DataFrame({'col': list(x.columns),
                               'imp': rfmodel.feature_importances_})
    rf_fea_imp = rf_fea_imp.sort_values('imp', ascending=False).reset_index(drop=True).iloc[:imp_num, :]
    rf_select_col = list(rf_fea_imp.col)
    return rf_fea_imp, rf_select_col


# 相关性可视化
def plot_corr(df, col_list, threshold=None, plt_size=None, is_annot=True):
    """
    df:数据集
    col_list:变量list集合
    threshold: 相关性设定的阈值
    plt_size:图纸尺寸
    is_annot:是否显示相关系数值

    return :相关性热力图
    """
    corr_df = df.loc[:, col_list].corr()
    plt.figure(figsize=plt_size)
    sns.heatmap(corr_df, annot=is_annot, cmap='rainbow', vmax=1, vmin=-1, mask=np.abs(corr_df) <= threshold)
    return plt.show()


# 相关性变量映射关系 
def corr_mapping(df, col_list, threshold=None):
    """
    df:数据集
    col_list:变量list集合
    threshold: 相关性设定的阈值

    return:强相关性变量之间的映射关系表
    """
    corr_df = df.loc[:, col_list].corr()
    col_a = []
    col_b = []
    corr_value = []
    for col, i in zip(col_list[:-1], range(1, len(col_list), 1)):
        high_corr_col = []
        high_corr_value = []
        corr_series = corr_df[col][i:]
        for i, j in zip(corr_series.index, corr_series.values):
            if abs(j) >= threshold:
                high_corr_col.append(i)
                high_corr_value.append(j)
        col_a.extend([col] * len(high_corr_col))
        col_b.extend(high_corr_col)
        corr_value.extend(high_corr_value)

    corr_map_df = pd.DataFrame({'col_A': col_a,
                                'col_B': col_b,
                                'corr': corr_value})
    return corr_map_df


# 相关性剔除
def forward_delete_corr(df, col_list, threshold=None):
    """
    df:数据集
    col_list:变量list集合
    threshold: 相关性设定的阈值

    return:相关性剔除后的变量
    """
    list_corr = col_list[:]
    for col in list_corr:
        corr = df.loc[:, list_corr].corr()[col]
        corr_index = [x for x in corr.index if x != col]
        corr_values = [x for x in corr.values if x != 1]
        for i, j in zip(corr_index, corr_values):
            if abs(j) >= threshold:
                list_corr.remove(i)
    return list_corr


# 显著性筛选,在筛选前需要做woe转换
def forward_delete_pvalue(x_train, y_train):
    """
    x_train -- x训练集
    y_train -- y训练集

    return :显著性筛选后的变量
    """
    col_list = list(x_train.columns)
    pvalues_col = []
    for col in col_list:
        pvalues_col.append(col)
        x_train2 = sm.add_constant(x_train.loc[:, pvalues_col])
        sm_lr = sm.Logit(y_train, x_train2)
        sm_lr = sm_lr.fit()
        for i, j in zip(sm_lr.pvalues.index[1:], sm_lr.pvalues.values[1:]):
            if j >= 0.05:
                pvalues_col.remove(i)

    x_new_train = x_train.loc[:, pvalues_col]
    x_new_train2 = sm.add_constant(x_new_train)
    lr = sm.Logit(y_train, x_new_train2)
    lr = lr.fit()
    print(lr.summary2())
    return pvalues_col, lr.summary2()


# 逻辑回归系数符号筛选,在筛选前需要做woe转换
def forward_delete_coef(x_train, y_train):
    """
    x_train -- x训练集 
    y_train -- y训练集

    return :
    coef_col回归系数符号筛选后的变量
    lr_coe：每个变量的系数值
    """
    col_list = list(x_train.columns)
    coef_col = []
    for col in col_list:
        coef_col.append(col)
        x_train2 = x_train.loc[:, coef_col]
        sk_lr = LogisticRegression(random_state=0).fit(x_train2, y_train)
        coef_df = pd.DataFrame({'col': coef_col, 'coef': sk_lr.coef_[0]})
        if coef_df[coef_df.coef < 0].shape[0] > 0:
            coef_col.remove(col)

    x_new_train = x_train.loc[:, coef_col]
    lr = LogisticRegression(random_state=0).fit(x_new_train, y_train)
    lr_coe = pd.DataFrame({'col': coef_col,
                           'coef': lr.coef_[0]})
    return coef_col, lr_coe


# 相关性剔除（考虑IV）
def forward_delete_corr_ivfirst(df, col_list, iv_rank, threshold=0.5):
    '''
    df: 数据集
    col_list: 变量list
    iv_rank: 变量的IV的list
    threshold: corr的筛选阈值

    return: 考虑了IV的大小之后，筛选出来的变量list
    '''

    def up_triangle(df, col_list, iv_rank, threshold=0.5):
        '''
        like above
        '''
        # initial
        list_corr = col_list[:]
        # 计算变量之间的corr并存表
        corr_df = df.loc[:, col_list].corr()
        # 遍历corr_df的所有corr
        for col in list_corr:
            for row in list_corr:
                corr = corr_df.loc[row, col]
                # 记录横纵变量的iv
                iv_col = (iv_rank[iv_rank['col'] == col]['iv']).values
                iv_row = (iv_rank[iv_rank['col'] == row]['iv']).values
                # 判断
                if corr > threshold and col != row:
                    if iv_col > iv_row:
                        list_corr.remove(row)
                        # print('delete %s %s > %s' %(row,iv_col,iv_row))
                    elif iv_col <= iv_row:
                        list_corr.remove(row)
                        # print('delete %s %s > %s' %(row,iv_col,iv_row))
                        break  # 如果删除了row，则没办法继续for循环，所以要break
        return list_corr

    once = up_triangle(df, col_list, iv_rank=iv_rank, threshold=threshold)
    twice = up_triangle(df, once, iv_rank=iv_rank, threshold=threshold)
    return twice


# 相关性剔除（考虑xgboost_imp or rf_imp）
def forward_delete_corr_impfirst(df, col_list, fea_imp, threshold=0.5):
    '''
    df: 数据集
    col_list: 变量list
    fea_imp: 变量的imp的list,
             可为xgboost:xg_fea_imp
             ramdomforest:rf_fea_imp
    threshold: corr的筛选阈值

    return: 考虑了imp的大小之后，筛选出来的变量list
    '''

    def up_triangle(df, col_list, fea_imp, threshold):
        '''
        like above
        '''
        # initial
        list_corr = col_list[:]
        # 计算变量之间的corr并存表
        corr_df = df.loc[:, col_list].corr()
        # 遍历corr_df的所有corr
        for col in list_corr:
            for row in list_corr:
                corr = corr_df.loc[row, col]
                # 记录横纵变量的iv
                imp_col = (fea_imp[fea_imp['col'] == col]['imp']).values
                imp_row = (fea_imp[fea_imp['col'] == row]['imp']).values
                # 判断
                if corr > threshold and col != row:
                    if imp_col > imp_row:
                        list_corr.remove(row)
                        # print('delete %s %s > %s' %(row,iv_col,iv_row))
                    elif imp_col <= imp_row:
                        list_corr.remove(row)
                        # print('delete %s %s > %s' %(row,iv_col,iv_row))
                        break  # 如果删除了row，则没办法继续for循环，所以要break
        return list_corr

    once = up_triangle(df, col_list, fea_imp, threshold)
    twice = up_triangle(df, once, fea_imp, threshold)
    return twice


# depth_first_search

def depth_first_search(x_train, y_train, x_test, y_test, col_list, col_initial, loop_num, length):
    """
    x_train: 训练集x
    y_train: 训练集target
    x_test: 测试集x
    y_test: 测试集target
    col_list: 参与搜索的变量list
    col_initial: 参与搜索的变量list的启动list
    loop_num: 希望搜索的次数
    length: 希望搜索的变量list的长度
    """
    j = 0
    coef = []
    intercept = []
    ks_list = []
    roc_list = []
    col_func = []
    # col_pvalue_delete_list = []
    # lr_list = []
    # col_corr_delete_list = []
    # coef_col_list = []
    # lr_coe_list = []

    while j < loop_num:
        b = ''
        col = col_initial
        # model_outside = LogisticRegression()
        while len(col) < length:
            # print(len(col))
            a = random.choice(col_list)
            # print(a)
            if b == a:
                # print('b == a')
                a = random.choice(col_list)
                # print(a)
                col.append(a)
                col = list(set(col))
                b = a
            else:
                # print('b <> a')
                col.append(a)
                col = list(set(col))
                b = a
            # print(col)

        # 用逻辑回归训练model,默认最大迭代100次，可能会超出限制，建议多设置一点
        model = LogisticRegression(max_iter=3000)
        model.fit(x_train[col], y_train)
        # y_pred用predict_proba，因为模型整体能力较弱，基本只有0.5一下的比率
        y_pred = model.predict_proba(x_test[col])[:, 1]
        # ROC
        roc = metrics.roc_auc_score(y_test, y_pred)
        roc_list.append(roc)
        # KS
        ks_max = model_evaluation.model_ks(y_test, y_pred)
        print(col, ks_max, roc)

        '''
        #筛选p-value
        col_pvalue_delete,lr = forward_delete_pvalue(df[col],df['dlq_flag'])
        col_pvalue_delete_list.append(col_pvalue_delete)
        lr_list.append(lr)

        #筛选corr
        col_corr_delete = forward_delete_corr(df,col_list,threshold=0.6)
        col_corr_delete_list.append(col_corr_delete)

        #筛选woe系数
        coef_col,lr_coe = forward_delete_coef(df[col],df['dlq_flag'])
        coef_col_list.append(list(set(lr_coe['col'])))
        lr_coe_list.append(list(set(lr_coe['coef'])))
        '''

        # 记录coef
        coef.append(model.coef_[0])
        # 记录intercept
        intercept.append(model.intercept_[0])
        # 记录ks_max
        ks_list.append(ks_max)
        # 记录col
        col_func.append(col)

        j = j + 1
        print(j)

    ks_col_list = pd.DataFrame({'col_list': col_func,
                                'ks_list': ks_list,
                                'ROC': roc_list,
                                'intercept': intercept,
                                'coef': coef,
                                #                            'col_pvalue_delete_list': col_pvalue_delete_list,
                                #                            'lr_list':lr_list,
                                #                            'col_corr_delete_list': col_corr_delete_list,
                                #                            'coef_col_list': coef_col_list,
                                #                            'lr_coe': lr_coe_list,
                                })
    return ks_col_list


# 在 main.py 或者 __init__.py 中创建主类 Yihui
import warnings
import pandas as pd
import numpy as np
from Yihui.yihui import Yihui

# 在主程序中使用 Yihui 类
if __name__ == "__main__":
    # ban FutureWarning
    warnings.filterwarnings('ignore')

    # generate data.csv
    with open("../Data/data_yinlian.csv", "r") as f:
        data = pd.read_csv(f)

    df = data.copy()
    df = df.drop(
        ['Unnamed: 0', 'APPLY_SERIAL_NO', 'apply_time', 'CUSTOMER_ID', 'CERT_ID', 'PHONE_NUMBER', 'CUSTOMER_NAME'],
        axis=1)

    yi = Yihui(df, 'dlq')

    print("Categorical Variables:", yi.get_categorical_variables())
    print("Numeric Variables:", yi.get_numeric_variables())
    print("Date Variables:", yi.get_date_variables())

    df = yi.dp_module.date_var_shift_binary(yi.get_date_variables(),replace=True)
    print(df['UPPA007_binary'])
#
#     print(yi.get_categorical_variables())
#     # yi.dp_module.fillna_cate_var(yi.get_categorical_variables(),'class','unkonwn')
#     # yi.eda_module.plot_default_cate(yi.get_categorical_variables(),plt_size=(100,100),plt_num=3,x=3,y=1)
#
#     yi.data = yi.data.drop(yi.get_categorical_variables(),axis = 1)
# )
#     # l = []
#     # for i in yi.get_numeric_variables():
#     #     if len(l) == 100:
#     #         yi.eda_module.plot_num_col(l,plt_type='box',plt_size=(100,100),plt_num=100,x=10,y=10)
#     #         l = []
#     #     else:
#     #         l.append(i)
#
#     yi.data = yi.dp_module.fillna_num_var(yi.get_numeric_variables(),fill_type='class',fill_class_num=-999)
#     # yi.dp_module.plot_bar_missing_var()
#
#     yi.binning_module.binning_num(yi.get_numeric_variables(),20,0)


import math
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy.stats import spearmanr


class BinningModule:

    def __init__(self, yihui_instance):
        self.yihui_instance = yihui_instance
        self.bin_df = None
        self.woe_list = None
        self.iv_df = None
        self.ks_df = None

    def binning_cate(self, col_list):
        df = self.yihui_instance.data
        target = self.yihui_instance.target

        total = df[target].count()
        bad = df[target].sum()
        good = total - bad

        all_odds = good * 1.0 / bad

        bin_df = []
        iv_value = []
        ks_value = []
        for col in col_list:
            d1 = df.groupby([col], as_index=True)
            d2 = pd.DataFrame()
            d2['min_bin'] = d1[col].min()
            d2['max_bin'] = d1[col].max()
            d2['total'] = d1[target].count()
            d2['totalrate'] = d2['total'] / total
            d2['bad'] = d1[target].sum()
            d2['badrate'] = d2['bad'] / d2['total']
            d2['good'] = d2['total'] - d2['bad']
            d2['goodrate'] = d2['good'] / d2['total']
            d2['badattr'] = d2['bad'] / bad
            d2['goodattr'] = d2['good'] / good
            d2['cumgoodrate'] = (d2['goodrate'] * d2['totalrate'] / (good / total)).cumsum()
            d2['cumbadrate'] = (d2['badrate'] * d2['totalrate'] / (bad / total)).cumsum()
            d2['odds'] = d2['good'] / d2['bad']
            GB_list = []
            for i in d2.odds:
                if i >= all_odds:
                    GB_index = str(round((i / all_odds) * 100, 0)) + str('G')
                else:
                    GB_index = str(round((all_odds / i) * 100, 0)) + str('B')
                GB_list.append(GB_index)
            d2['GB_index'] = GB_list
            # WOEi = log( (BADi/BADtotal)/(GOODi/GODDtotal) )
            d2['woe'] = np.log(d2['badattr'] / d2['goodattr'])
            # IVi = ( (BADi/BADtotal) - (GOODi/GODDtotal) ) * WOEi
            d2['bin_iv'] = (d2['badattr'] - d2['goodattr']) * d2['woe']
            # IV = sum(IVi)
            d2['IV'] = d2['bin_iv'].sum()
            d2['ks'] = np.abs(d2['badattr'] - d2['goodattr'])
            iv = d2['bin_iv'].sum().round(3)
            ks = d2['ks'].max()
            print('var:{}'.format(col))
            print('IV:{}'.format(iv))
            print('KS:{}'.format(ks))
            bin_df.append(d2)
            iv_value.append(iv)
            ks_value.append(ks)
        self.bin_df = bin_df
        return bin_df, iv_value, ks_value

    # 数值型变量的分箱

    # 先用卡方分箱输出变量的分割点
    def __split_data(self, df, col, split_num):
        """
        df: 原始数据集
        col:需要分箱的变量
        split_num:分割点的数量
        """
        df2 = df.copy()
        count = df2.shape[0]  # 总样本数
        n = math.floor(count / split_num)  # 按照分割点数目等分后每组的样本数
        split_index = [i * n for i in range(1, split_num)]  # 分割点的索引
        values = sorted(list(df2[col]))  # 对变量的值从小到大进行排序
        split_value = [values[i] for i in split_index]  # 分割点对应的value
        split_value = sorted(list(set(split_value)))  # 分割点的value去重排序
        return split_value

    def __assign_group(self, x, split_bin):
        """
        x:变量的value
        split_bin:__split_data得出的分割点list
        """
        n = len(split_bin)
        if x <= min(split_bin):
            return min(split_bin)  # 如果x小于分割点的最小值，则x映射为分割点的最小值
        elif x > max(split_bin):  # 如果x大于分割点的最大值，则x映射为分割点的最大值
            return 10e10
        else:
            for i in range(n - 1):
                if split_bin[i] < x <= split_bin[i + 1]:  # 如果x在两个分割点之间，则x映射为分割点较大的值
                    return split_bin[i + 1]

    def __bin_bad_rate(self, df, col, target, grantRateIndicator=0):
        """
        df:原始数据集
        col:原始变量/变量映射后的字段
        target:目标变量的字段
        grantRateIndicator:是否输出总体的违约率
        """
        total = df.groupby([col])[target].count()
        bad = df.groupby([col])[target].sum()
        total_df = pd.DataFrame({'total': total})
        bad_df = pd.DataFrame({'bad': bad})
        regroup = pd.merge(total_df, bad_df, left_index=True, right_index=True, how='left')
        regroup = regroup.reset_index()
        regroup['bad_rate'] = regroup['bad'] / regroup['total']  # 计算根据col分组后每组的违约率
        dict_bad = dict(zip(regroup[col], regroup['bad_rate']))  # 转为字典形式
        if grantRateIndicator == 0:
            return (dict_bad, regroup)
        total_all = df.shape[0]
        bad_all = df[target].sum()
        all_bad_rate = bad_all / total_all  # 计算总体的违约率
        return (dict_bad, regroup, all_bad_rate)

    # def __cal_chi2(df,all_bad_rate):
    #     """
    #     df:__bin_bad_rate得出的regroup
    #     all_bad_rate:__bin_bad_rate得出的总体违约率
    #     """
    #     df2 = df.copy()
    #     df2['expected'] = df2['total']*all_bad_rate # 计算每组的坏用户期望数量
    #     combined = zip(df2['expected'],df2['bad']) # 遍历每组的坏用户期望数量和实际数量
    #     chi = [(i[0]-i[1])**2/i[0] for i in combined] # 计算每组的卡方值
    #     chi2 = sum(chi) # 计算总的卡方值
    #     return chi2

    def __cal_chi2(self, df, all_bad_rate):
        """
        df: __bin_bad_rate得出的regroup
        all_bad_rate: __bin_bad_rate得出的总体违约率
        """
        df2 = df.copy()
        df2['expected'] = df2['total'] * all_bad_rate  # 计算每组的坏用户期望数量

        # Add a small constant to avoid division by zero
        epsilon = 1e-10
        df2['expected'] = df2['expected'].apply(lambda x: x + epsilon if x == 0 else x)

        combined = zip(df2['expected'], df2['bad'])  # 遍历每组的坏用户期望数量和实际数量
        chi = [(i[0] - i[1]) ** 2 / i[0] for i in combined]  # 计算每组的卡方值
        chi2 = sum(chi)  # 计算总的卡方值
        return chi2

    def __assign_bin(self, x, cutoffpoints):
        """
        x:变量的value
        cutoffpoints:分箱的切割点
        """
        bin_num = len(cutoffpoints) + 1  # 箱体个数
        if x <= cutoffpoints[0]:  # 如果x小于最小的cutoff点，则映射为Bin 0
            return 'Bin 0'
        elif x > cutoffpoints[-1]:  # 如果x大于最大的cutoff点，则映射为Bin(bin_num-1)
            return 'Bin {}'.format(bin_num - 1)
        else:
            for i in range(0, bin_num - 1):
                if cutoffpoints[i] < x <= cutoffpoints[i + 1]:  # 如果x在两个cutoff点之间，则x映射为Bin(i+1)
                    return 'Bin {}'.format(i + 1)

    def __chi_merge(self, df, col, target, max_bin=5, min_binpct=0):
        col_unique = sorted(list(set(df[col])))  # 变量的唯一值并排序
        n = len(col_unique)  # 变量唯一值得个数
        df2 = df.copy()
        if n > 100:  # 如果变量的唯一值数目超过100，则将通过__split_data和__assign_group将x映射为split对应的value
            split_col = self.__split_data(df2, col, 100)  # 通过这个目的将变量的唯一值数目人为设定为100
            df2['col_map'] = df2[col].map(lambda x: self.__assign_group(x, split_col))
        else:
            df2['col_map'] = df2[col]  # 变量的唯一值数目没有超过100，则不用做映射
        # 生成dict_bad,regroup,all_bad_rate的元组
        (dict_bad, regroup, all_bad_rate) = self.__bin_bad_rate(df2, 'col_map', target, grantRateIndicator=1)
        col_map_unique = sorted(list(set(df2['col_map'])))  # 对变量映射后的value进行去重排序
        group_interval = [[i] for i in col_map_unique]  # 对col_map_unique中每个值创建list并存储在group_interval中

        while (len(group_interval) > max_bin):  # 当group_interval的长度大于max_bin时，执行while循环
            chi_list = []
            for i in range(len(group_interval) - 1):
                temp_group = group_interval[i] + group_interval[i + 1]  # temp_group 为生成的区间,list形式，例如[1,3]
                chi_df = regroup[regroup['col_map'].isin(temp_group)]
                chi_value = self.__cal_chi2(chi_df, all_bad_rate)  # 计算每一对相邻区间的卡方值
                chi_list.append(chi_value)
            best_combined = chi_list.index(min(chi_list))  # 最小的卡方值的索引
            # 将卡方值最小的一对区间进行合并
            group_interval[best_combined] = group_interval[best_combined] + group_interval[best_combined + 1]
            # 删除合并前的右区间
            group_interval.remove(group_interval[best_combined + 1])
            # 对合并后每个区间进行排序
        group_interval = [sorted(i) for i in group_interval]
        # cutoff点为每个区间的最大值
        cutoffpoints = [max(i) for i in group_interval[:-1]]

        # 检查是否有箱只有好样本或者只有坏样本
        df2['col_map_bin'] = df2['col_map'].apply(lambda x: self.__assign_bin(x, cutoffpoints))  # 将col_map映射为对应的区间Bin
        # 计算每个区间的违约率
        (dict_bad, regroup) = self.__bin_bad_rate(df2, 'col_map_bin', target)
        # 计算最小和最大的违约率
        [min_bad_rate, max_bad_rate] = [min(dict_bad.values()), max(dict_bad.values())]
        # 当最小的违约率等于0，说明区间内只有好样本，当最大的违约率等于1，说明区间内只有坏样本
        while min_bad_rate == 0 or max_bad_rate == 1:
            bad01_index = regroup[regroup['bad_rate'].isin([0, 1])].col_map_bin.tolist()  # 违约率为1或0的区间
            bad01_bin = bad01_index[0]
            if bad01_bin == max(regroup.col_map_bin):
                cutoffpoints = cutoffpoints[:-1]  # 当bad01_bin是最大的区间时，删除最大的cutoff点
            elif bad01_bin == min(regroup.col_map_bin):
                cutoffpoints = cutoffpoints[1:]  # 当bad01_bin是最小的区间时，删除最小的cutoff点
            else:
                bad01_bin_index = list(regroup.col_map_bin).index(bad01_bin)  # 找出bad01_bin的索引
                prev_bin = list(regroup.col_map_bin)[bad01_bin_index - 1]  # bad01_bin前一个区间
                df3 = df2[df2.col_map_bin.isin([prev_bin, bad01_bin])]
                (dict_bad, regroup1) = self.__bin_bad_rate(df3, 'col_map_bin', target)
                chi1 = self.__cal_chi2(regroup1, all_bad_rate)  # 计算前一个区间和bad01_bin的卡方值
                later_bin = list(regroup.col_map_bin)[bad01_bin_index + 1]  # bin01_bin的后一个区间
                df4 = df2[df2.col_map_bin.isin([later_bin, bad01_bin])]
                (dict_bad, regroup2) = self.__bin_bad_rate(df4, 'col_map_bin', target)
                chi2 = self.__cal_chi2(regroup2, all_bad_rate)  # 计算后一个区间和bad01_bin的卡方值
                if chi1 < chi2:  # 当chi1<chi2时,删除前一个区间对应的cutoff点
                    cutoffpoints.remove(cutoffpoints[bad01_bin_index - 1])
                else:  # 当chi1>=chi2时,删除bin01对应的cutoff点
                    cutoffpoints.remove(cutoffpoints[bad01_bin_index])
            df2['col_map_bin'] = df2['col_map'].apply(lambda x: self.__assign_bin(x, cutoffpoints))
            (dict_bad, regroup) = self.__bin_bad_rate(df2, 'col_map_bin', target)
            # 重新将col_map映射至区间，并计算最小和最大的违约率，直达不再出现违约率为0或1的情况，循环停止
            [min_bad_rate, max_bad_rate] = [min(dict_bad.values()), max(dict_bad.values())]

        # 检查分箱后的最小占比
        if min_binpct > 0:
            group_values = df2['col_map'].apply(lambda x: self.__assign_bin(x, cutoffpoints))
            df2['col_map_bin'] = group_values  # 将col_map映射为对应的区间Bin
            group_df = group_values.value_counts().to_frame()
            group_df['bin_pct'] = group_df['col_map'] / n  # 计算每个区间的占比
            min_pct = group_df.bin_pct.min()  # 得出最小的区间占比
            while min_pct < min_binpct and len(cutoffpoints) > 2:  # 当最小的区间占比小于min_pct且cutoff点的个数大于2，执行循环
                # 下面的逻辑基本与“检验是否有箱体只有好/坏样本”的一致
                min_pct_index = group_df[group_df.bin_pct == min_pct].index.tolist()
                min_pct_bin = min_pct_index[0]
                if min_pct_bin == max(group_df.index):
                    cutoffpoints = cutoffpoints[:-1]
                elif min_pct_bin == min(group_df.index):
                    cutoffpoints = cutoffpoints[1:]
                else:
                    minpct_bin_index = list(group_df.index).index(min_pct_bin)
                    prev_pct_bin = list(group_df.index)[minpct_bin_index - 1]
                    df5 = df2[df2['col_map_bin'].isin([min_pct_bin, prev_pct_bin])]
                    (dict_bad, regroup3) = self.__bin_bad_rate(df5, 'col_map_bin', target)
                    chi3 = self.__cal_chi2(regroup3, all_bad_rate)
                    later_pct_bin = list(group_df.index)[minpct_bin_index + 1]
                    df6 = df2[df2['col_map_bin'].isin([min_pct_bin, later_pct_bin])]
                    (dict_bad, regroup4) = self.__bin_bad_rate(df6, 'col_map_bin', target)
                    chi4 = self.__cal_chi2(regroup4, all_bad_rate)
                    if chi3 < chi4:
                        cutoffpoints.remove(cutoffpoints[minpct_bin_index - 1])
                    else:
                        cutoffpoints.remove(cutoffpoints[minpct_bin_index])
        return cutoffpoints

    # 数值型变量的分箱
    def binning_num(self, col_list, max_bin=None, min_binpct=None, method='ChiMerge', n=10, leaf_stop_percent=0.05):
        """
        df:数据集
        target:目标变量的字段名
        col_list:变量list集合
        max_bin:最大的分箱个数
        min_binpct:区间内样本所占总体的最小比
        method:分箱方式（freq,count,ChiMerge(default),cart)
        leaf_stop_percent:叶子节点占比，作为停止条件，默认5%()
    
        return:
        bin_df :list形式，里面存储每个变量的分箱结果
        iv_value:list形式，里面存储每个变量的IV值
        """
        df = self.yihui_instance.data.copy()
        target = self.yihui_instance.target

        total = df[target].count()
        bad = df[target].sum()
        good = total - bad
        all_odds = good / bad
        inf = float('inf')
        ninf = float('-inf')
        bin_df = []
        iv_value = []

        for col in col_list:
            if method == 'freq':  # 等频分箱
                bucket = pd.qcut(df[col], n, duplicates='drop')
            elif method == 'count':  # 等距分箱
                bucket = pd.cut(df[col], n)
            # elif method == 'cart':  #决策树cart
            #     cut = sc.binning_function.get_cart_bincut(df, col, target, leaf_stop_percent=leaf_stop_percent)
            #     bucket = pd.cut(df[col],cut)
            elif method == 'monotonic':
                r = 0
                while np.abs(r) < 1:
                    d_1 = pd.DataFrame({'col': df[col],
                                        'target': df[target],
                                        'bin': pd.qcut(df[col], n)})
                    d_2 = d_1.groupby('bin', as_index=True)
                    r, p = spearmanr(d_2['col'].mean(), d_2['target'].mean())
                    n = n + 1
                bucket = pd.qcut(df[col], n - 1)
                print(bucket)
            elif method == 'ChiMerge':  # 卡方
                cut = self.__chi_merge(df, col, target, max_bin=max_bin, min_binpct=min_binpct)
                cut.insert(0, ninf)
                cut.append(inf)
                bucket = pd.cut(df[col], cut)
            d1 = df.groupby(bucket)
            d2 = pd.DataFrame()
            d2['min_bin'] = d1[col].min()
            d2['max_bin'] = d1[col].max()
            d2['total'] = d1[target].count()
            d2['totalrate'] = d2['total'] / total
            d2['bad'] = d1[target].sum()
            d2['badrate'] = d2['bad'] / d2['total']
            d2['good'] = d2['total'] - d2['bad']
            d2['goodrate'] = d2['good'] / d2['total']
            d2['badattr'] = d2['bad'] / bad
            d2['goodattr'] = (d2['total'] - d2['bad']) / good
            d2['cumgoodrate'] = (d2['goodrate'] * d2['totalrate'] / (good / total)).cumsum()
            d2['cumbadrate'] = (d2['badrate'] * d2['totalrate'] / (bad / total)).cumsum()
            d2['odds'] = d2['good'] / d2['bad']
            GB_list = []
            for i in d2.odds:
                if i >= all_odds:
                    GB_index = str(round((i / all_odds) * 100, 0)) + str('G')
                else:
                    GB_index = str(round((all_odds / i) * 100, 0)) + str('B')
                GB_list.append(GB_index)
            d2['GB_index'] = GB_list
            d2['woe'] = np.log(d2['badattr'] / d2['goodattr'])
            d2['bin_iv'] = (d2['badattr'] - d2['goodattr']) * d2['woe']
            d2['IV'] = d2['bin_iv'].sum()
            iv = d2['bin_iv'].sum().round(3)
            print('变量名:{}'.format(col))
            print('IV:{}'.format(iv))
            print('\t')
            bin_df.append(d2)
            iv_value.append(iv)
            d2.reset_index(inplace=True)
        return bin_df, iv_value

    # 数值型变量的iv明细表
    def iv_num(self, col_list, **kwargs):
        """
        df:数据集
        target:目标变量的字段名
        col_list:变量list集合
        max_bin:最大的分箱个数
        min_binpct:区间内样本所占总体的最小比

        return :变量的iv明细表
        """
        max_bin = kwargs.get('max_bin', None)
        min_binpct = kwargs.get('min_binpct', None)
        method = kwargs.get('method', None)

        df = self.yihui_instance.data.copy()
        target = self.yihui_instance.target

        bin_df, iv_value = self.binning_num(col_list, max_bin=max_bin, min_binpct=min_binpct, method='ChiMerge')
        iv_df = pd.DataFrame({'col': col_list,
                              'iv': iv_value})
        self.iv_df = iv_df
        return iv_df


    # 数值型变量的分箱（手动分箱）
    def binning_num_manual(self, df, target, col, cut):
        """
        df:数据集
        target:目标变量的字段名
        col:变量名
        cut:断点list
    
        return:
        bin_df :list形式，里面存储变量的分箱结果
        iv_value:list形式，里面存储每个变量的IV值
        """
        total = df[target].count()
        bad = df[target].sum()
        good = total - bad
        all_odds = good / bad
        inf = float('inf')
        ninf = float('-inf')
        bin_df = []
        iv_value = []

        col = col
        cut = cut
        cut.insert(0, ninf)
        cut.append(inf)
        bucket = pd.cut(df[col], cut)
        d1 = df.groupby(bucket)
        d2 = pd.DataFrame()
        d2['min_bin'] = d1[col].min()
        d2['max_bin'] = d1[col].max()
        d2['total'] = d1[target].count()
        d2['totalrate'] = d2['total'] / total
        d2['bad'] = d1[target].sum()
        d2['badrate'] = d2['bad'] / d2['total']
        d2['good'] = d2['total'] - d2['bad']
        d2['goodrate'] = d2['good'] / d2['total']
        d2['badattr'] = d2['bad'] / bad
        d2['goodattr'] = (d2['total'] - d2['bad']) / good
        d2['cumgoodrate'] = (d2['goodrate'] * d2['totalrate'] / (good / total)).cumsum()
        d2['cumbadrate'] = (d2['badrate'] * d2['totalrate'] / (bad / total)).cumsum()
        d2['odds'] = d2['good'] / d2['bad']
        GB_list = []
        for i in d2.odds:
            if i >= all_odds:
                GB_index = str(round((i / all_odds) * 100, 0)) + str('G')
            else:
                GB_index = str(round((all_odds / i) * 100, 0)) + str('B')
            GB_list.append(GB_index)
        d2['GB_index'] = GB_list
        d2['woe'] = np.log(d2['badattr'] / d2['goodattr'])
        d2['bin_iv'] = (d2['badattr'] - d2['goodattr']) * d2['woe']
        d2['IV'] = d2['bin_iv'].sum()
        iv = d2['bin_iv'].sum().round(3)
        print('变量名:{}'.format(col))
        print('IV:{}'.format(iv))
        print('\t')
        bin_df.append(d2)
        iv_value.append(iv)
        d2.reset_index(inplace=True)

        return bin_df, iv_value



    # 自定义分箱
    def binning_self(self, col, cut=None, right_border=True):
        """
        df: 数据集
        col:分箱的单个变量名
        cut:划分区间的list
        right_border：设定左开右闭、左闭右开
    
        return:
        bin_df: df形式，单个变量的分箱结果
        iv_value: 单个变量的iv
        """
        df = self.yihui_instance.data.copy()
        target = self.yihui_instance.target

        total = df[target].count()
        bad = df[target].sum()
        good = total - bad
        all_odds = good / bad
        bucket = pd.cut(df[col], cut, right=right_border)
        d1 = df.groupby(bucket)
        d2 = pd.DataFrame()
        d2['min_bin'] = d1[col].min()
        d2['max_bin'] = d1[col].max()
        d2['total'] = d1[target].count()
        d2['totalrate'] = d2['total'] / total
        d2['bad'] = d1[target].sum()
        d2['badrate'] = d2['bad'] / d2['total']
        d2['good'] = d2['total'] - d2['bad']
        d2['goodrate'] = d2['good'] / d2['total']
        d2['badattr'] = d2['bad'] / bad
        d2['goodattr'] = (d2['total'] - d2['bad']) / good
        d2['odds'] = d2['good'] / d2['bad']
        GB_list = []
        for i in d2.odds:
            if i >= all_odds:
                GB_index = str(round((i / all_odds) * 100, 0)) + str('G')
            else:
                GB_index = str(round((all_odds / i) * 100, 0)) + str('B')
            GB_list.append(GB_index)
        d2['GB_index'] = GB_list
        d2['woe'] = np.log(d2['badattr'] / d2['goodattr'])
        d2['bin_iv'] = (d2['badattr'] - d2['goodattr']) * d2['woe']
        d2['IV'] = d2['bin_iv'].sum()
        iv_value = d2['bin_iv'].sum().round(3)
        print('变量名:{}'.format(col))
        print('IV:{}'.format(iv_value))
        bin_df = d2.copy()
        return bin_df, iv_value

    # 变量分箱结果的检查

    # woe的可视化
    def plot_woe(self, hspace=0.4, wspace=0.4, plt_size=None, plt_num=None, x=None, y=None):
        """
        bin_df:list形式，里面存储每个变量的分箱结果
        hspace :子图之间的间隔(y轴方向)
        wspace :子图之间的间隔(x轴方向)
        plt_size :图纸的尺寸
        plt_num :子图的数量
        x :子图矩阵中一行子图的数量
        y :子图矩阵中一列子图的数量
    
        return :每个变量的woe变化趋势图
        """
        bin_df = self.bin_df.copy()
        plt.figure(figsize=plt_size)
        plt.subplots_adjust(hspace=hspace, wspace=wspace)
        for i, df in zip(range(1, plt_num + 1, 1), bin_df):
            col_name = df.columns.values[0]
            df = df.reset_index()
            plt.subplot(x, y, i)
            plt.title(col_name, color='cornflowerblue')
            sns.barplot(data=df, x=col_name, y='woe')
            plt.xlabel('')
            plt.xticks(ticks=[])  # rotation=30 #ticks = [] 去掉x轴坐标
        return plt.show()

    # 检验woe是否单调
    def woe_monoton(self):
        """
        bin_df:list形式，里面存储每个变量的分箱结果
    
        return :
        woe_notmonoton_col :woe没有呈单调变化的变量，list形式
        woe_judge_df :df形式，每个变量的检验结果
        """
        bin_df = self.bin_df.copy()
        woe_notmonoton_col = []
        col_list = []
        woe_judge = []
        for woe_df in bin_df:
            col_name = woe_df.index.name
            woe_list = list(woe_df.woe)
            if woe_df.shape[0] == 2:
                # print('{}是否单调: True'.format(col_name))
                col_list.append(col_name)
                woe_judge.append('True')
            else:
                woe_not_monoton = [(woe_list[i] < woe_list[i + 1] and woe_list[i] < woe_list[i - 1]) or (
                            woe_list[i] > woe_list[i + 1] and woe_list[i] > woe_list[i - 1]) for i in
                                   range(1, len(woe_list) - 1, 1)]
                if True in woe_not_monoton:
                    # print('{}是否单调: False'.format(col_name))
                    woe_notmonoton_col.append(col_name)
                    col_list.append(col_name)
                    woe_judge.append('False')
                else:
                    # print('{}是否单调: True'.format(col_name))
                    col_list.append(col_name)
                    woe_judge.append('True')
        woe_judge_df = pd.DataFrame({'col': col_list,
                                     'judge_monoton': woe_judge})
        return woe_notmonoton_col, woe_judge_df

    # 检查某个区间的woe是否大于1
    def woe_large(self):
        """
        bin_df:list形式，里面存储每个变量的分箱结果
    
        return:
        woe_large_col: 某个区间woe大于1的变量，list集合
        woe_judge_df :df形式，每个变量的检验结果
        """
        bin_df = self.bin_df.copy()

        woe_large_col = []
        col_list = []
        woe_judge = []
        for woe_df in bin_df:
            col_name = woe_df.index.name
            woe_list = list(woe_df.woe)
            woe_large = list(filter(lambda x: x >= 1, woe_list))
            if len(woe_large) > 0:
                col_list.append(col_name)
                woe_judge.append('True')
                woe_large_col.append(col_name)
            else:
                col_list.append(col_name)
                woe_judge.append('False')
        woe_judge_df = pd.DataFrame({'col': col_list,
                                     'judge_large': woe_judge})
        return woe_large_col, woe_judge_df



import random
import matplotlib.pyplot as plt
# plt.style.use('science')

# 变量筛选
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
from sklearn import metrics
from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from xgboost import XGBClassifier


class VarSelectModule:

    def __init__(self, yihui_instance):
        self.yihui_instance = yihui_instance
        self.xg_fea_imp = None
        self.rf_fea_imp = None

    # xgboost筛选变量
    def select_xgboost(self, col_list, imp_num=None):
        """
        df:数据集
        target:目标变量的字段名
        imp_num:筛选变量的个数

        return:
        xg_fea_imp:变量的特征重要性
        xg_select_col:筛选出的变量
        """
        target = self.yihui_instance.target

        x = self.yihui_instance.data[col_list].copy()
        y = self.yihui_instance.data[target]

        xgmodel = XGBClassifier(random_state=0)
        xgmodel = xgmodel.fit(x, y, eval_metric='auc')
        xg_fea_imp = pd.DataFrame({'col': list(x.columns),
                                   'imp': xgmodel.feature_importances_})
        xg_fea_imp_rank = xg_fea_imp.sort_values('imp', ascending=False).reset_index(drop=True).iloc[:imp_num, :]
        xg_select_col = list(xg_fea_imp.col)

        self.xg_fea_imp = xg_fea_imp

        return xg_fea_imp, xg_fea_imp_rank, xg_select_col

    # 随机森林筛选变量
    def select_rf(self, col_list, imp_num=None):
        """
        df:数据集
        target:目标变量的字段名
        imp_num:筛选变量的个数

        return:
        rf_fea_imp:变量的特征重要性
        rf_select_col:筛选出的变量
        """
        target = self.yihui_instance.target

        x = self.yihui_instance.data[col_list].copy()
        y = self.yihui_instance.data[target]

        rfmodel = RandomForestClassifier(random_state=0)
        rfmodel = rfmodel.fit(x, y)
        rf_fea_imp = pd.DataFrame({'col': list(x.columns),
                                   'imp': rfmodel.feature_importances_})
        rf_fea_imp = rf_fea_imp.sort_values('imp', ascending=False).reset_index(drop=True).iloc[:imp_num, :]
        rf_select_col = list(rf_fea_imp.col)

        self.rf_fea_imp = rf_fea_imp

        return rf_fea_imp, rf_select_col

    # 相关性可视化
    def plot_corr(self, col_list, threshold=None, mask_direction='lt', plt_size=None, is_annot=False):
        """
        df:数据集
        col_list:变量list集合
        threshold: 相关性设定的阈值
        plt_size:图纸尺寸
        is_annot:是否显示相关系数值

        return :相关性热力图
        """
        df = self.yihui_instance.data.copy()

        corr_df = df.loc[:, col_list].corr()
        plt.figure(figsize=plt_size)
        if mask_direction == 'gt':
            sns.heatmap(corr_df, annot=is_annot, cmap='rainbow', vmax=1, vmin=-1, mask=np.abs(corr_df) >= threshold)
        if mask_direction == 'lt':
            sns.heatmap(corr_df, annot=is_annot, cmap='rainbow', vmax=1, vmin=-1, mask=np.abs(corr_df) <= threshold)
        return plt.show()

    # 相关性变量映射关系
    def corr_mapping(self, col_list, threshold=None, direction='lt'):
        """
        df:数据集
        col_list:变量list集合
        threshold: 相关性设定的阈值

        return:强相关性变量之间的映射关系表
        """
        df = self.yihui_instance.data.copy()

        corr_df = df.loc[:, col_list].corr()
        col_a = []
        col_b = []
        corr_value = []
        for col, i in zip(col_list[:-1], range(1, len(col_list), 1)):
            high_corr_col = []
            high_corr_value = []
            corr_series = corr_df[col][i:]
            for i, j in zip(corr_series.index, corr_series.values):
                if threshold == None:
                    print('Error: corr_mapping threshold is None')
                    quit()
                if direction == 'gt':
                    if abs(j) >= threshold:
                        high_corr_col.append(i)
                        high_corr_value.append(j)
                if direction == 'lt':
                    if abs(j) < threshold:
                        high_corr_col.append(i)
                        high_corr_value.append(j)
            col_a.extend([col] * len(high_corr_col))
            col_b.extend(high_corr_col)
            corr_value.extend(high_corr_value)

        corr_map_df = pd.DataFrame({'col_A': col_a,
                                    'col_B': col_b,
                                    'corr': corr_value})
    # 相关性剔除
        return corr_map_df

    def forward_delete_corr(self, col_list, threshold=None):
        """
        df:数据集
        col_list:变量list集合
        threshold: 相关性设定的阈值

        return:相关性剔除后的变量
        """
        df = self.yihui_instance.data.copy()

        list_corr = col_list[:]
        for col in list_corr:
            corr = df.loc[:, list_corr].corr()[col]
            corr_index = [x for x in corr.index if x != col]
            corr_values = [x for x in corr.values if x != 1]
            for i, j in zip(corr_index, corr_values):
                if abs(j) >= threshold:
                    list_corr.remove(i)
                    print(i,j)
        return list_corr

    # 相关性剔除（考虑IV）
    def forward_delete_corr_ivfirst(self, col_list, threshold=0.5):
        '''
        df: 数据集
        col_list: 变量list
        iv_rank: 变量的IV的list
        threshold: corr的筛选阈值

        return: 考虑了IV的大小之后，筛选出来的变量list
        '''

        global iv_rank

        def up_triangle(df, col_list, iv_rank, threshold=0.5):
            '''
            like above
            '''
            # initial
            list_corr = col_list[:]
            # 计算变量之间的corr并存表
            corr_df = df.loc[:, col_list].corr()
            # 遍历corr_df的所有corr
            for col in list_corr:
                for row in list_corr:
                    corr = corr_df.loc[row, col]
                    if iv_rank is not None:
                        # 记录横纵变量的iv
                        iv_col = (iv_rank[iv_rank['col'] == col]['iv']).values
                        iv_row = (iv_rank[iv_rank['col'] == row]['iv']).values
                    # elif fea_imp is not None:
                    #     # 记录横纵变量的iv
                    #     coll = (fea_imp[fea_imp['col'] == col]['imp']).values
                    #     roww = (fea_imp[fea_imp['col'] == row]['imp']).values
                    # 判断
                    if corr > threshold and col != row:
                        if iv_col > iv_row:
                            list_corr.remove(row)
                            # print('delete %s %s > %s' %(row,iv_col,iv_row))
                        elif iv_col <= iv_row:
                            list_corr.remove(row)
                            # print('delete %s %s > %s' %(row,iv_col,iv_row))
                            break  # 如果删除了row，则没办法继续for循环，所以要break
            return list_corr

        df = self.yihui_instance.data.copy()

        if self.yihui_instance.binning_module.iv_df is not None:
            print(self.yihui_instance.binning_module.iv_df)
            iv_rank = self.yihui_instance.binning_module.iv_df.sort_values(by='iv',ascending=False)
        elif self.yihui_instance.binning_module.iv_df is None:
            _, iv_rank = self.yihui_instance.binning_module.iv_num(
                col_list, max_bin=20, min_binpct=0, method='ChiMerge')

        # iv_rank = iv_rank.sort_values(by='iv',ascending=False)

        once = up_triangle(df, col_list, iv_rank=iv_rank, threshold=threshold)
        twice = up_triangle(df, once, iv_rank=iv_rank, threshold=threshold)
        return twice



    # 相关性剔除（考虑xgboost_imp or rf_imp）
    def forward_delete_corr_impfirst(self, col_list, type, threshold=0.5):
        '''
        df: 数据集
        col_list: 变量list
        fea_imp: 变量的imp的list,
                 可为xgboost:xg_fea_imp
                 ramdomforest:rf_fea_imp
        threshold: corr的筛选阈值

        return: 考虑了imp的大小之后，筛选出来的变量list
        '''

        def up_triangle(df, col_list, fea_imp, threshold):
            '''
            like above
            '''
            # initial
            list_corr = col_list[:]
            # 计算变量之间的corr并存表
            corr_df = df.loc[:, col_list].corr()
            # 遍历corr_df的所有corr
            for col in list_corr:
                for row in list_corr:
                    corr = corr_df.loc[row, col]
                    # 记录横纵变量的iv
                    imp_col = (fea_imp[fea_imp['col'] == col]['imp']).values
                    imp_row = (fea_imp[fea_imp['col'] == row]['imp']).values
                    # 判断
                    if corr > threshold and col != row:
                        if imp_col > imp_row:
                            list_corr.remove(row)
                            # print('delete %s %s > %s' %(row,iv_col,iv_row))
                        elif imp_col <= imp_row:
                            list_corr.remove(row)
                            # print('delete %s %s > %s' %(row,iv_col,iv_row))
                            break  # 如果删除了row，则没办法继续for循环，所以要break
            return list_corr

        df = self.yihui_instance.data.copy()

        if type == 'xgboost':
            if self.xg_fea_imp is not None:
                fea_imp = self.xg_fea_imp.sort_values(by='imp',ascending=False)
            elif self.xg_fea_imp is None:
               fea_imp,_,_ = self.select_xgboost(col_list)
               fea_imp = fea_imp.sort_values(by='imp',ascending=False)

        once = up_triangle(df, col_list, fea_imp, threshold)
        twice = up_triangle(df, once, fea_imp, threshold)
        return twice

    # depth_first_search

    def depth_first_search(x_train, y_train, x_test, y_test, col_list, col_initial, loop_num, length):
        """
        x_train: 训练集x
        y_train: 训练集target
        x_test: 测试集x
        y_test: 测试集target
        col_list: 参与搜索的变量list
        col_initial: 参与搜索的变量list的启动list
        loop_num: 希望搜索的次数
        length: 希望搜索的变量list的长度
        """
        j = 0
        coef = []
        intercept = []
        ks_list = []
        roc_list = []
        col_func = []
        # col_pvalue_delete_list = []
        # lr_list = []
        # col_corr_delete_list = []
        # coef_col_list = []
        # lr_coe_list = []

        while j < loop_num:
            b = ''
            col = col_initial
            # model_outside = LogisticRegression()
            while len(col) < length:
                # print(len(col))
                a = random.choice(col_list)
                # print(a)
                if b == a:
                    # print('b == a')
                    a = random.choice(col_list)
                    # print(a)
                    col.append(a)
                    col = list(set(col))
                    b = a
                else:
                    # print('b <> a')
                    col.append(a)
                    col = list(set(col))
                    b = a
                # print(col)

            # 用逻辑回归训练model,默认最大迭代100次，可能会超出限制，建议多设置一点
            model = LogisticRegression(max_iter=3000)
            model.fit(x_train[col], y_train)
            # y_pred用predict_proba，因为模型整体能力较弱，基本只有0.5一下的比率
            y_pred = model.predict_proba(x_test[col])[:, 1]
            # ROC
            roc = metrics.roc_auc_score(y_test, y_pred)
            roc_list.append(roc)
            # KS
            ks_max = model_evaluation.model_ks(y_test, y_pred)
            print(col, ks_max, roc)

            '''
            #筛选p-value
            col_pvalue_delete,lr = forward_delete_pvalue(df[col],df['dlq_flag'])
            col_pvalue_delete_list.append(col_pvalue_delete)
            lr_list.append(lr)

            #筛选corr
            col_corr_delete = forward_delete_corr(df,col_list,threshold=0.6)
            col_corr_delete_list.append(col_corr_delete)

            #筛选woe系数
            coef_col,lr_coe = forward_delete_coef(df[col],df['dlq_flag'])
            coef_col_list.append(list(set(lr_coe['col'])))
            lr_coe_list.append(list(set(lr_coe['coef'])))
            '''

            # 记录coef
            coef.append(model.coef_[0])
            # 记录intercept
            intercept.append(model.intercept_[0])
            # 记录ks_max
            ks_list.append(ks_max)
            # 记录col
            col_func.append(col)

            j = j + 1
            print(j)

        ks_col_list = pd.DataFrame({'col_list': col_func,
                                    'ks_list': ks_list,
                                    'ROC': roc_list,
                                    'intercept': intercept,
                                    'coef': coef,
                                    #                            'col_pvalue_delete_list': col_pvalue_delete_list,
                                    #                            'lr_list':lr_list,
                                    #                            'col_corr_delete_list': col_corr_delete_list,
                                    #                            'coef_col_list': coef_col_list,
                                    #                            'lr_coe': lr_coe_list,
                                    })
        return ks_col_list














    # 显著性筛选,在筛选前需要做woe转换
    def forward_delete_pvalue(x_train, y_train):
        """
        x_train -- x训练集
        y_train -- y训练集

        return :显著性筛选后的变量
        """
        col_list = list(x_train.columns)
        pvalues_col = []
        for col in col_list:
            pvalues_col.append(col)
            x_train2 = sm.add_constant(x_train.loc[:, pvalues_col])
            sm_lr = sm.Logit(y_train, x_train2)
            sm_lr = sm_lr.fit()
            for i, j in zip(sm_lr.pvalues.index[1:], sm_lr.pvalues.values[1:]):
                if j >= 0.05:
                    pvalues_col.remove(i)

        x_new_train = x_train.loc[:, pvalues_col]
        x_new_train2 = sm.add_constant(x_new_train)
        lr = sm.Logit(y_train, x_new_train2)
        lr = lr.fit()
        print(lr.summary2())
        return pvalues_col, lr.summary2()

    # 逻辑回归系数符号筛选,在筛选前需要做woe转换
    def forward_delete_coef(self, x_train, y_train):
        """
        x_train -- x训练集
        y_train -- y训练集

        return :
        coef_col回归系数符号筛选后的变量
        lr_coe：每个变量的系数值
        """
        col_list = list(x_train.columns)
        coef_col = []
        for col in col_list:
            coef_col.append(col)
            x_train2 = x_train.loc[:, coef_col]
            sk_lr = LogisticRegression(random_state=0).fit(x_train2, y_train)
            coef_df = pd.DataFrame({'col': coef_col, 'coef': sk_lr.coef_[0]})
            if coef_df[coef_df.coef < 0].shape[0] > 0:
                coef_col.remove(col)

        x_new_train = x_train.loc[:, coef_col]
        lr = LogisticRegression(random_state=0).fit(x_new_train, y_train)
        lr_coe = pd.DataFrame({'col': coef_col,
                               'coef': lr.coef_[0]})
        return coef_col, lr_coe





import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.ensemble import RandomForestRegressor

class DataProcessingModule:
    def __init__(self, yihui_instance):
        self.yihui_instance = yihui_instance

    def __missing_var_cal(self):
        """
        calculate var missing pct
        """
        total = self.yihui_instance.data.shape[0]
        missing_count = self.yihui_instance.data.isnull().sum()
        missing_pct = missing_count/total
        missing_data = pd.DataFrame({
            'index': self.yihui_instance.data.columns.tolist(),
            'total_obs': total,
            'missing_count': missing_count,
            'missing_pct': missing_pct
        })
        return missing_data

    def __missing_obs_cal(self):
        """
        calculate obs missing pct
        """
        total = len(self.yihui_instance.data.columns)
        missing_count = self.yihui_instance.data.isnull().sum(axis=1)
        missing_pct = missing_count / total
        missing_data = pd.DataFrame({
            'index': self.yihui_instance.data.index.tolist(),
            'total_obs': total,
            'missing_count': missing_count,
            'missing_pct': missing_pct
        })
        print(missing_data)
        return missing_data



    # 所有变量缺失值分布图
    def plot_bar_missing_var(self, plt_size=None):
        """
        plt_size: plot chart size: (10, 10)

        return: bar chart of missing variables
        """
        missing_data = self.__missing_var_cal()
        print('plot_bar')
        print(missing_data)
        if plt_size is not None:
            plt.figure(figsize=plt_size)
        else:
            plt.figure()

        plt.rcParams['font.sans-serif'] = ['Microsoft YaHei']
        plt.rcParams['axes.unicode_minus'] = False

        x = missing_data['index']
        y = missing_data['missing_pct']

        plt.bar(x, y, color='hotpink', edgecolor='k', alpha=0.8)
        plt.title('各列缺失率')
        plt.ylabel('缺失率')
        plt.xlabel('列名')
        plt.xticks(rotation=45)

        plt.show()

    # TODO:
    #* 单个样本缺失值分布图
    # def plot_bar_missing_obs(self):
        


    # 缺失值填充（类别型变量）
    def fillna_cate_var(self, col_list, fill_type=None, fill_str=None):
        """
        data:数据集
        col_list:变量list集合
        fill_type: 填充方式：众数/当做一个类别

        return :填充后的数据集
        """
        data2 = self.yihui_instance.data.copy()
        for col in col_list:
            if fill_type == 'class':
                data2[col] = data2[col].fillna(fill_str)
            if fill_type == 'mode':
                data2[col] = data2[col].fillna(data2[col].mode()[0])
        return data2

    # 数值型变量的填充
    # 针对缺失率在5%以下的变量用中位数填充
    # 缺失率在5%--15%的变量用随机森林填充,可先对缺失率较低的变量先用中位数填充，在用没有缺失的样本来对变量作随机森林填充
    # 缺失率超过15%的变量建议当作一个类别
    def fillna_num_var(self, col_list, fill_type=None, fill_class_num=None, filled_data=None):
        """
        data:数据集
        col_list:变量list集合
        fill_type:填充方式：0/中位数/当作一个类别/随机森林
        fill_class_num:用特殊数值填充，当填充方式为当作一个类别时使用
        filled_data :已填充好的数据集，当填充方式为随机森林时使用

        return:已填充好的数据集
        """
        data2 = self.yihui_instance.data.copy()
        for col in col_list:
            if fill_type == '0':
                data2[col] = data2[col].fillna(0)
            if fill_type == 'median':
                data2[col] = data2[col].fillna(data2[col].median())
            if fill_type == 'class':
                data2[col] = data2[col].fillna(fill_class_num)
            if fill_type == 'rf':
                rf_data = pd.concat([data2[col], filled_data], axis=1)
                known = rf_data[rf_data[col].notnull()]
                unknown = rf_data[rf_data[col].isnull()]
                x_train = known.drop([col], axis=1)
                y_train = known[col]
                x_pre = unknown.drop([col], axis=1)
                rf = RandomForestRegressor(random_state=0)
                rf.fit(x_train, y_train)
                y_pre = rf.predict(x_pre)
                data2.loc[data2[col].isnull(), col] = y_pre
        return data2

    # 缺失值剔除（单个变量）
    def delete_missing_var(self, threshold=None):
        """
        data:数据集
        threshold:缺失率删除的阈值

        return :删除缺失后的数据集
        """
        data2 = self.yihui_instance.data.copy()
        missing_data = self.__missing_var_cal()
        missing_col_num = missing_data[missing_data.missing_pct >= threshold].shape[0]
        missing_col = list(missing_data[missing_data.missing_pct >= threshold].index)
        data2 = data2.drop(missing_col, axis=1)
        print('缺失率超过{}的变量个数为{}'.format(threshold, missing_col_num))
        return data2

    def delete_missing_obs(self, threshold=None):
        """
        删除包含超过阈值数量的缺失值的 observation

        Parameters:
        - threshold: 允许的最大缺失变量数量

        Returns:
        删除缺失后的数据集
        """
        data2 = self.yihui_instance.data.copy()
        # 计算每个 observation 中缺失值的数量
        missing_data = self.__missing_obs_cal()

        # 找到缺失值数量超过阈值的 observation 的索引
        if threshold >= 1:
            obs_to_remove = missing_data[missing_data['missing_count'] >= threshold].index
        else:
            obs_to_remove = missing_data[missing_data['missing_pct'] >= threshold].index
        data2 = data2.drop(obs_to_remove, axis=0)
        print('含有超过{}个缺失值的样本数量为{}'.format(threshold, len(obs_to_remove)))
        return data2

    # 常变量/同值化处理
    def const_delete(self, threshold=0.9):
        """
        删除常变量/同值化处理

        Parameters:
        - threshold: 同值化处理的阈值，默认为 0.9

        Returns:
        删除常变量/同值化处理后的数据集
        """
        # 计算每一列中唯一值的比例
        unique_ratio = self.yihui_instance.data.nunique() / len(self.yihui_instance.data)

        # 找到同值比例超过阈值的列
        const_columns = unique_ratio[unique_ratio >= threshold].index

        # 删除常变量/同值化处理后的数据集
        data_after_const_delete = self.yihui_instance.data.drop(columns=const_columns)

        print('删除常变量/同值化处理后的变量个数为{},名字为{}'.format(len(const_columns),const_columns))
        print(data_after_const_delete.shape[1])
        return data_after_const_delete

    # 缺失目标变量删除
    def target_missing_delete(self):
        """
        删除目标变量为空的观测

        Returns:
        删除缺失目标变量后的数据集
        """
        if self.yihui_instance.target is not None:
            data_without_missing_target = self.yihui_instance.data.dropna(subset=[self.yihui_instance.target])
            missing_target_count = len(self.yihui_instance.data) - len(data_without_missing_target)
            print('删除目标变量缺失的观测数: {}'.format(missing_target_count))
            return data_without_missing_target
        else:
            print('未指定目标变量，无法执行删除操作。')

    # 日期变量转换为二进制变量
    def date_var_shift_binary(self, col_list, replace=False):
        """
        将指定的日期型变量转换成二进制变量

        Parameters:
        - date_variable: str, 指定的日期型变量名称

        Returns:
        - data_with_binary: pd.DataFrame, 包含二进制变量的数据集
        """

        data = self.yihui_instance.data.copy()
        # 创建一个二进制变量，表示日期是否存在
        for col in col_list:
            data[col + '_binary'] = 0
            data.loc[~data[col].isnull(), col + '_binary'] = 1
            if not replace:
                # 如果需要保留原始日期变量，可以选择删除原始日期列
                data.drop(columns=[col], inplace=True)

        return data


import matplotlib.pyplot as plt

# plt.style.use('science')
import numpy as np
from numpy import unique
from numpy import where
from sklearn.cluster import AffinityPropagation  # 亲和力传播
from sklearn.cluster import AgglomerativeClustering  # 聚合
from sklearn.cluster import Birch  # BIRCH
from sklearn.cluster import DBSCAN  # DBSCAN
from sklearn.cluster import KMeans  # K-Means
from sklearn.cluster import MeanShift  # 均值漂移
from sklearn.cluster import MiniBatchKMeans  # Mini-Batch K-Means
from sklearn.cluster import OPTICS  # OPTICS
from sklearn.cluster import SpectralClustering  # 光谱聚合
from sklearn.mixture import GaussianMixture  # 高斯模糊

from tqdm import tqdm

class ClusterMuodule():
    def __init__(self, yihui_instance):
        self.yihui_instance = yihui_instance

    # 亲和力传播
    # cluster_AffinityPropagation(df,['v1','v2'],damping = 0.9)
    def cluster_AffinityPropagation(self, col_list, damping=0.9):
        """
        亲和力传播包括找到一组最能概括数据的范例。
        我们设计了一种名为“亲和传播”的方法，它作为两对数据点之间相似度的输入度量。在数据点之间交换实值消息，直到一组高质量的范例和相应的群集逐渐出现。
        它是通过 AffinityPropagation 类实现的.
        要调整的主要配置是将“ 阻尼 ”设置为0.5到1，甚至可能是“首选项”。
        """
        # 定义模型
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        model = AffinityPropagation(damping=damping)
        # 匹配模型
        print('fitting model')
        model.fit(X)
        # 为每个示例分配一个集群
        d1 = model.predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters, desc="plotting"):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # 聚合聚类
    # cluster_AgglomerativeClustering(df,['v1','v2'],n_clusters = 2)
    def cluster_AgglomerativeClustering(self, col_list, n_clusters=2):
        """
        聚合聚类涉及合并示例，直到达到所需的群集数量为止。
        它是层次聚类方法的更广泛类的一部分，通过AgglomerationClustering类实现的。
        主要配置是“ n _ clusters ”集，这是对数据中的群集数量的估计。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model1 = AgglomerativeClustering(n_clusters=n_clusters)
        # 模型拟合与聚类预测
        print('fitting model')
        d1 = model1.fit_predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # BIRCH聚类
    # cluster_Birch(df,['v1','v2'],threshold = 0.01, n_clusters = 2)
    def cluster_Birch(self, col_list, threshold=0.01, n_clusters=2):
        """
        BIRCH 聚类（ BIRCH 是平衡迭代减少的缩写，聚类使用层次结构)
        包括构造一个树状结构，从中提取聚类质心。
        BIRCH 递增地和动态地群集传入的多维度量数据点，以尝试利用可用资源（即可用内存和时间约束）产生最佳质量的聚类。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        model = Birch(threshold=threshold, n_clusters=n_clusters)
        # 适配模型
        print('fitting model')
        model.fit(X)
        # 为每个示例分配一个集群
        d1 = model.predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # DBSCAN聚类
    # cluster_DBSCAN(df,col_list,eps=0.30, min_samples=9)
    def cluster_DBSCAN(self, col_list, eps=0.30, min_samples=9):
        """
        DBSCAN 聚类（其中 DBSCAN 是基于密度的空间聚类的噪声应用程序）
        涉及在域中寻找高密度区域，并将其周围的特征空间区域扩展为群集。
        我们提出了新的聚类算法 DBSCAN 依赖于基于密度的概念的集群设计，以发现任意形状的集群。
        DBSCAN 只需要一个输入参数，并支持用户为其确定适当的值
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model = DBSCAN(eps=eps, min_samples=min_samples)
        # 模型拟合与聚类预测
        print('fitting model')
        d1 = model.fit_predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # K-Means聚类
    # cluster_KMeans(df,col_list,n_clusters=10)
    def cluster_KMeans(self, col_list, n_clusters=2):
        """
        K-均值聚类可以是最常见的聚类算法，并涉及向群集分配示例，以尽量减少每个群集内的方差。
        本文的主要目的是描述一种基于样本将 N 维种群划分为 k 个集合的过程。
        这个叫做“ K-均值”的过程似乎给出了在类内方差意义上相当有效的分区。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model = KMeans(n_clusters=n_clusters)
        # 适配模型
        print('fitting model')
        model.fit(X)
        # 为每个示例分配一个集群
        d1 = model.predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # Mini-Batch K-Means聚类
    # cluster_MiniBatchKMeans(df,col_list,n_clusters = 10)
    def cluster_MiniBatchKMeans(self, col_list, n_clusters=2):
        """
        Mini-Batch K-均值是 K-均值的修改版本。
        它使用小批量的样本而不是整个数据集对群集质心进行更新，这可以使大数据集的更新速度更快，并且可能对统计噪声更健壮。
        我们建议使用 k-均值聚类的迷你批量优化。
        与经典批处理算法相比，这降低了计算成本的数量级，同时提供了比在线随机梯度下降更好的解决方案。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model = MiniBatchKMeans(n_clusters=n_clusters)
        # 适配模型
        print('fitting model')
        model.fit(X)
        # 为每个示例分配一个集群
        d1 = model.predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # 均值飘逸聚类
    # cluster_MeanShift(df,col_list)
    def cluster_MeanShift(self, col_list):
        """
        均值漂移聚类涉及到根据特征空间中的实例密度来寻找和调整质心。
        对离散数据证明了递推平均移位程序收敛到最接近驻点的基础密度函数，从而证明了它在检测密度模式中的应用。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model = MeanShift()
        # 适配模型
        print('fitting model')
        model.fit(X)
        # 为每个示例分配一个集群
        d1 = model.predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # OPTICS聚类
    # cluster_OPTICS(df,col_list,eps=0.8,min_samples=10)
    def cluster_OPTICS(self, col_list, eps=0.8, min_samples=10):
        """
        OPTICS 聚类（ OPTICS 短于订购点数以标识聚类结构）是上述 DBSCAN 的修改版本。
        我们为聚类分析引入了一种新的算法，它不会显式地生成一个数据集的聚类；
        而是创建表示其基于密度的聚类结构的数据库的增强排序。
        此群集排序包含相当于密度聚类的信息，该信息对应于范围广泛的参数设置。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        model = OPTICS(eps=eps, min_samples=min_samples)
        # 为每个示例分配一个集群
        print('fitting model')
        d1 = model.fit_predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # 光谱聚类
    # cluster_SpectralClustering(df,col_list,n_clusters=10)
    def cluster_SpectralClustering(self, col_list, n_clusters=2):
        """
        光谱聚类是一类通用的聚类方法，取自线性线性代数。
        最近在许多领域出现的一个有希望的替代方案是使用聚类的光谱方法。
        这里，使用从点之间的距离导出的矩阵的顶部特征向量。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        print('fitting model')
        model = SpectralClustering(n_clusters=n_clusters)
        # 为每个示例分配一个集群
        d1 = model.fit_predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters, desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()

    # 高斯混合聚类
    # cluster_GaussianMixture(df,col_list,n_components=10)
    def cluster_GaussianMixture(self, col_list, n_components=2):
        """
        高斯混合模型总结了一个多变量概率密度函数，顾名思义就是混合了高斯概率分布。
        """
        # 定义数据集
        df = self.yihui_instance.data
        X = np.array(df[col_list])
        # 定义模型
        print('fitting model')
        model = GaussianMixture(n_components=n_components)
        # 为每个示例分配一个集群
        d1 = model.fit_predict(X)
        # 检索唯一群集
        clusters = unique(d1)
        # 为每个群集的样本创建散点图
        for cluster in tqdm(clusters,desc='plotting'):
            # 获取此群集的示例的行索引
            row_ix = where(d1 == cluster)
            # 创建这些样本的散布
            plt.scatter(X[row_ix, 0], X[row_ix, 1])
            # 绘制散点图
        plt.show()



# 在 main.py 或者 __init__.py 中创建主类 Yihui
import warnings
import pandas as pd
import numpy as np
from Yihui.yihui import Yihui

# 在主程序中使用 Yihui 类
if __name__ == "__main__":
    # ban FutureWarning
    warnings.filterwarnings('ignore')

    # generate data.csv
    with open("../Data/data.csv", "r") as f:
        data = pd.read_csv(f)
    data['customer_no'] = str(data['customer_no'])
    data['v101'] = np.random.choice(['A', 'B', 'C', 'D', 'E', 'F'], size=len(data))

    # loading Titanic dataset
    # data = sns.load_dataset('titanic')
    # data.head()

    # Create Yihui Class
    yihui_project = Yihui(data, 'dlq_flag')
    print(yihui_project.data.head())

    print("Categorical Variables:", yihui_project.get_categorical_variables())
    print("Numeric Variables:", yihui_project.get_numeric_variables())
    print("Date Variables:", yihui_project.get_date_variables())

    # 直接访问 Yihui 类的属性获取字符型和数值型变量的名字
    categorical_vars_list = yihui_project.get_categorical_variables()
    numeric_vars_list = yihui_project.get_numeric_variables()

    eda_result = yihui_project.eda_module.auto_eda_simple()
    print(eda_result)


    # ### eda 阶段
    #
    # # 使用ydata_profiling 自动生成eda报告
    # # 根据dataset数据量大小，生成报告的时间会不同。建议慎重操作。
    # yihui_project.eda_module.auto_eda_profiling()
    #
    # # 手动查看变量分布情况
    # yihui_project.eda_module.plot_num_col(numeric_vars_list,plt_type='hist',plt_size=(100,100),plt_num=100,x=10,y=10)
    yihui_project.eda_module.plot_num_col(numeric_vars_list,plt_type='box',plt_size=(100,100),plt_num=100,x=10,y=10)
    # yihui_project.eda_module.plot_cate_var(categorical_vars_list,plt_size=(100,100),plt_num=100,x=10,y=10)
    #
    # # 数值型变量的违约率分析
    # yihui_project.eda_module.plot_default_num(numeric_vars_list,q=10,plt_size=(100,100),plt_num=100,x=10,y=10)
    #
    # # 类别型变量的违约率分析
    # yihui_project.eda_module.plot_default_cate(categorical_vars_list,plt_size=(10,10),plt_num=1,x=1,y=1)
    #

    ### data processing 阶段

    # 所有变量缺失值分布图
    # print(yihui_project.dp_module.plot_bar_missing_var())

    # 使用 '0','median','class','rf'
    yihui_project.data = yihui_project.dp_module.fillna_num_var(numeric_vars_list, fill_type='0')
    #
    # yihui_project.data = yihui_project.dp_module.fillna_cate_var(categorical_vars_list, fill_type='class', fill_str='missing')
    # yihui_project.data = yihui_project.dp_module.fillna_cate_var(categorical_vars_list, fill_type='mode')
    #
    # # 缺失值剔除
    # yihui_project.data = yihui_project.dp_module.delete_missing_var(threshold=0.2)
    # yihui_project.data = yihui_project.dp_module.delete_missing_obs(threshold=5)

    # 常变量/同值化处理
    yihui_project.data = yihui_project.dp_module.const_delete(threshold=0.5)
    # print(yihui_project.dp_module.const_delete(threshold=0.9).columns)
    print(yihui_project.get_numeric_variables())

    yihui_project.data = yihui_project.dp_module.target_missing_delete()
    # 再检查一下缺失值
    # yihui_project.dp_module.plot_bar_missing_var()

    # # cluster 阶段

    # yihui_project.cluster_module.cluster_AffinityPropagation(['v1','v2']) # 非常慢，慎重
    # yihui_project.cluster_module.cluster_Birch(['v1','v2'])
    # yihui_project.cluster_module.cluster_GaussianMixture(['v2','v3'])
    # yihui_project.cluster_module.cluster_KMeans(['v3','v4'])


    # # binning 阶段
    # yihui_project.binning_module.binning_cate(categorical_vars_list)

    # print(yihui_project.binning_module.bin_df)

    # yihui_project.binning_module.binning_num(yihui_project.get_numeric_variables(), max_bin=20, min_binpct=0, method='ChiMerge')
    # yihui_project.binning_module.binning_num(['v1', 'v2', 'v3', 'v4'], max_bin=20, min_binpct=0, method='ChiMerge')
    # yihui_project.binning_module.iv_num(['v1', 'v2', 'v3', 'v4'], max_bin=20, min_binpct=0)
    # print(yihui_project.binning_module.iv_df)
    # yihui_project.binning_module.plot_woe(plt_size=(50,50),plt_num=4,x=2,y=2)


    # # var select 阶段
    # xg_fea_imp,xg_fea_imp_rank, xg_select_col = yihui_project.var_select_module.select_xgboost(yihui_project.get_numeric_variables())
    # print(xg_fea_imp)
    # print(xg_fea_imp['imp'].sum())
    # print(xg_select_col)

    # rf_fea_imp,  rf_select_col = yihui_project.var_select_module.select_rf(yihui_project.get_numeric_variables())
    # print(rf_fea_imp)
    # print(rf_fea_imp['imp'].sum())
    # print(rf_select_col)

    # yihui_project.var_select_module.plot_corr(yihui_project.get_numeric_variables(),0.5,'gt')
    #
    # corr_map_df = yihui_project.var_select_module.corr_mapping(yihui_project.get_numeric_variables(), threshold=0.5)
    # print(corr_map_df)
    #
    # list_corr = yihui_project.var_select_module.forward_delete_corr(yihui_project.get_numeric_variables(),0.5)
    # print(list_corr)
    # yihui_project.var_select_module.plot_corr(list_corr)
    #
    twice = yihui_project.var_select_module.forward_delete_corr_ivfirst(['v1','v2','v3','v4'],threshold=0.5)
    print(twice)

    twice = yihui_project.var_select_module.forward_delete_corr_impfirst(['v1','v2','v3','v4'],type='xgboost',threshold=0.5)
    print(twice)


import math
import numpy as np
import pandas as pd
from scipy.stats import chi2
from scipy.stats import spearmanr


# 01 计算IV
def iv_count(data, var, target, weight):
    """ 计算iv值
    Args:
        data: DataFrame，拟操作的数据集
        var: String，拟计算IV值的变量名称
        target: String，Y列名称
    Returns:
        IV值， float
    """
    # 获取变量的唯一值
    value_list = set(list(np.unique(data[var])))
    iv = 0
    # 计算每个唯一值的错误率
    data_bad = pd.Series(data[data[target] == 1][var].values, index=data[data[target] == 1].index)
    data_good = pd.Series(data[data[target] == 0][var].values, index=data[data[target] == 0].index)
    len_bad = len(data_bad)
    len_good = len(data_good)
    # 遍历变量的唯一值
    for value in value_list:
        # 判断是否某类是否为0，避免出现无穷小值和无穷大值
        if sum(data_bad == value) == 0:
            bad_rate = 1 / len_bad
        else:
            bad_rate = sum(data_bad == value) / len_bad
        if sum(data_good == value) == 0:
            good_rate = 1 / len_good
        else:
            good_rate = sum(data_good == value) / len_good
        # 计算iv值
        iv += (good_rate - bad_rate) * math.log(good_rate / bad_rate, 2)
        # print(value,iv)
    # 返回iv值
    return iv



# 02 基于CART算法的最优分箱代码实现


def get_var_median(data, var):
    """ 得到指定连续变量的所有元素的中位数列表
    Args:
        data: DataFrame，拟操作的数据集
        var: String，拟分箱的连续型变量名称
    Returns:
        关于连续变量的所有元素的中位列表，List
    """
    var_value_list = list(np.unique(data[var]))
    var_median_list = []
    for i in range(len(var_value_list) - 1):
        var_median = (var_value_list[i] + var_value_list[i + 1]) / 2
        var_median_list.append(var_median)
    return var_median_list


def calculate_gini(y):
    """ 计算基尼指数
    Args:
        y: Array，待计算数据的target，即0和1的数组
    Returns:
        基尼指数，float
    """
    # 将数组转化为列表
    y = y.tolist()
    probs = [y.count(i) / len(y) for i in np.unique(y)]
    gini = sum([p * (1 - p) for p in probs])
    return gini


def get_cart_split_point(data, var, target, min_sample):
    """ 获得最优的二值划分点（即基尼指数下降最大的点）
    Args:
        data: DataFrame，拟操作的数据集
        var: String，拟分箱的连续型变量名称
        target: String，Y列名称
        min_sample: int，分箱的最小数据样本，也就是数据量至少达到多少才需要去分箱，一般作用在开头或者结尾处的分箱点

    Returns:
        BestSplit_Point: 返回本次迭代的最优划分点，float
        BestSplit_Position: 返回最优划分点的位置，最左边为0，最右边为1，float
    """

    # 初始化
    Gini = calculate_gini(data[target].values)
    Best_Gini = 0.0
    BestSplit_Point = -99999
    BestSplit_Position = 0.0
    median_list = get_var_median(data, var)  # 获取当前数据集指定元素的所有中位数列表

    # 遍历中位数列表
    for i in range(len(median_list)):
        left = data[data[var] < median_list[i]]
        right = data[data[var] > median_list[i]]

        # 如果切分后的数据量少于指定阈值，跳出本次分箱计算
        if len(left) < min_sample or len(right) < min_sample:
            continue

        Left_Gini = calculate_gini(left[target].values)
        Right_Gini = calculate_gini(right[target].values)
        Left_Ratio = len(left) / len(data)
        Right_Ratio = len(right) / len(data)

        Temp_Gini = Gini - (Left_Gini * Left_Ratio + Right_Gini * Right_Ratio)
        if Temp_Gini > Best_Gini:
            Best_Gini = Temp_Gini
            BestSplit_Point = median_list[i]
            # 获取切分点的位置，最左边为0，最右边为1
            if len(median_list) > 1:
                BestSplit_Position = i / (len(median_list) - 1)
            else:
                BestSplit_Position = i / len(len(median_list))
        else:
            continue
    Gini = Gini - Best_Gini
    # print("最优切分点：", BestSplit_Point)
    return BestSplit_Point, BestSplit_Position


def get_cart_bincut(data, var, target, leaf_stop_percent=0.05):
    """ 计算最优分箱切分点
    Args:
        data: DataFrame，拟操作的数据集
        var: String，拟分箱的连续型变量名称
        target: String，Y列名称
        leaf_stop_percent: 叶子节点占比，作为停止条件，默认5%

    Returns:
        best_bincut: 最优的切分点列表，List
    """
    min_sample = len(data) * leaf_stop_percent
    best_bincut = []

    def cutting_data(data, var, target, min_sample, best_bincut):
        split_point, position = get_cart_split_point(data, var, target, min_sample)

        if split_point != -99999:
            best_bincut.append(split_point)

        # 根据最优切分点切分数据集，并对切分后的数据集递归计算切分点，直到满足停止条件
        # print("本次分箱的值域范围为{0} ~ {1}".format(data[var].min(), data[var].max()))
        left = data[data[var] < split_point]
        right = data[data[var] > split_point]

        # 当切分后的数据集仍大于最小数据样本要求，则继续切分
        if len(left) >= min_sample and position not in [0.0, 1.0]:
            cutting_data(left, var, target, min_sample, best_bincut)
        else:
            pass
        if len(right) >= min_sample and position not in [0.0, 1.0]:
            cutting_data(right, var, target, min_sample, best_bincut)
        else:
            pass
        return best_bincut

    best_bincut = cutting_data(data, var, target, min_sample, best_bincut)

    # 把切分点补上头尾
    best_bincut.append(data[var].min())
    best_bincut.append(data[var].max())
    best_bincut_set = set(best_bincut)
    best_bincut = list(best_bincut_set)

    best_bincut.remove(data[var].min())
    best_bincut.append(data[var].min() - 1)
    # 排序切分点
    best_bincut.sort()

    return best_bincut


# 03 基于卡方检验的最优分箱代码实现


def calculate_chi(freq_array):
    """ 计算卡方值
    Args:
        freq_array: Array，待计算卡方值的二维数组，频数统计结果
    Returns:
        卡方值，float
    """
    # 检查是否为二维数组
    assert (freq_array.ndim == 2)

    # 计算每列的频数之和
    col_nums = freq_array.sum(axis=0)
    # 计算每行的频数之和
    row_nums = freq_array.sum(axis=1)
    # 计算总频数
    nums = freq_array.sum()
    # 计算期望频数
    E_nums = np.ones(freq_array.shape) * col_nums / nums
    E_nums = (E_nums.T * row_nums).T
    # 计算卡方值
    tmp_v = (freq_array - E_nums) ** 2 / E_nums
    # 如果期望频数为0，则计算结果记为0
    tmp_v[E_nums == 0] = 0
    chi_v = tmp_v.sum()
    return chi_v


def get_chimerge_bincut(data, var, target, max_group=None, chi_threshold=None):
    """ 计算卡方分箱的最优分箱点
    Args:
        data: DataFrame，待计算卡方分箱最优切分点列表的数据集
        var: 待计算的连续型变量名称
        target: 待计算的目标列Y的名称
        max_group: 最大的分箱数量（因为卡方分箱实际上是合并箱体的过程，需要限制下最大可以保留的分箱数量）
        chi_threshold: 卡方阈值，如果没有指定max_group，我们默认选择类别数量-1，置信度95%来设置阈值

    Returns:
        最优切分点列表，List
    """

    '''
    如果不知道卡方阈值怎么取，可以生成卡方表来看看，代码如下：  
    import pandas as pd
    import numpy as np
    from scipy.stats import chi2
    p = [0.995, 0.99, 0.975, 0.95, 0.9, 0.5, 0.1, 0.05, 0.025, 0.01, 0.005]
    pd.DataFrame(np.array([chi2.isf(p, df=i) for i in range(1,10)]), columns=p, index=list(range(1,10)))
    '''

    freq_df = pd.crosstab(index=data[var], columns=data[target])
    # 转化为二维数组
    freq_array = freq_df.values

    # 初始化箱体，每个元素单独一组
    best_bincut = freq_df.index.values

    # 初始化阈值 chi_threshold，如果没有指定 chi_threshold，则默认选择target数量-1，置信度95%来设置阈值
    if max_group is None:
        if chi_threshold is None:
            chi_threshold = chi2.isf(0.05, df=freq_array.shape[-1])

    # 开始迭代
    while True:
        min_chi = None
        min_idx = None
        for i in range(len(freq_array) - 1):
            # 两两计算相邻两组的卡方值，得到最小卡方值的两组
            v = calculate_chi(freq_array[i: i + 2])
            if min_chi is None or min_chi > v:
                min_chi = v
                min_idx = i

        # 是否继续迭代条件判断
        # 条件1：当前箱体数仍大于 最大分箱数量阈值
        # 条件2：当前最小卡方值仍小于制定卡方阈值
        if (max_group is not None and max_group < len(freq_array)) or (
                chi_threshold is not None and min_chi < chi_threshold):
            tmp = freq_array[min_idx] + freq_array[min_idx + 1]
            freq_array[min_idx] = tmp
            freq_array = np.delete(freq_array, min_idx + 1, 0)
            best_bincut = np.delete(best_bincut, min_idx + 1, 0)
        else:
            break

    # 把切分点补上头尾
    best_bincut = best_bincut.tolist()
    best_bincut.append(data[var].min())
    best_bincut.append(data[var].max())
    best_bincut_set = set(best_bincut)
    best_bincut = list(best_bincut_set)

    best_bincut.remove(data[var].min())
    best_bincut.append(data[var].min() - 1)
    # 排序切分点
    best_bincut.sort()

    return best_bincut



# 04 基于最优KS的最优分箱代码实现

def get_maxks_split_point(data, var, target, min_sample=0.05):
    """ 计算KS值
    Args:
        data: DataFrame，待计算卡方分箱最优切分点列表的数据集
        var: 待计算的连续型变量名称
        target: 待计算的目标列Y的名称
        min_sample: int，分箱的最小数据样本，也就是数据量至少达到多少才需要去分箱，一般作用在开头或者结尾处的分箱点
    Returns:
        ks_v: KS值，float
        BestSplit_Point: 返回本次迭代的最优划分点，float
        BestSplit_Position: 返回最优划分点的位置，最左边为0，最右边为1，float
    """
    if len(data) < min_sample:
        ks_v, BestSplit_Point, BestSplit_Position = 0, -9999, 0.0
    else:
        # 计算每个组的卡方分箱点
        freq_df = pd.crosstab(index=data[var], columns=data[target])
        freq_array = freq_df.values
        if freq_array.shape[1] == 1:  # 如果某一组只有一个枚举值，如0或1，则数组形状会有问题，跳出本次计算
            # tt = np.zeros(freq_array.shape).T
            # freq_array = np.insert(freq_array, 0, values=tt, axis=1)
            ks_v, BestSplit_Point, BestSplit_Position = 0, -99999, 0.0
        else:
            bincut = freq_df.index.values
            tmp = freq_array.cumsum(axis=0) / (np.ones(freq_array.shape) * freq_array.sum(axis=0).T)
            tmp_abs = abs(tmp.T[0] - tmp.T[1])
            ks_v = tmp_abs.max()
            BestSplit_Point = bincut[tmp_abs.tolist().index(ks_v)]
            BestSplit_Position = tmp_abs.tolist().index(ks_v) / max(len(bincut) - 1, 1)

    return ks_v, BestSplit_Point, BestSplit_Position


def get_bestks_bincut(data, var, target, leaf_stop_percent=0.05):
    """ 计算最优分箱切分点
    Args:
        data: DataFrame，拟操作的数据集
        var: String，拟分箱的连续型变量名称
        target: String，Y列名称
        leaf_stop_percent: 叶子节点占比，作为停止条件，默认5%

    Returns:
        best_bincut: 最优的切分点列表，List
    """
    min_sample = len(data) * leaf_stop_percent
    best_bincut = []

    def cutting_data(data, var, target, min_sample, best_bincut):
        ks, split_point, position = get_maxks_split_point(data, var, target, min_sample)

        if split_point != -99999:
            best_bincut.append(split_point)

        # 根据最优切分点切分数据集，并对切分后的数据集递归计算切分点，直到满足停止条件
        # print("本次分箱的值域范围为{0} ~ {1}".format(data[var].min(), data[var].max()))
        left = data[data[var] < split_point]
        right = data[data[var] > split_point]

        # 当切分后的数据集仍大于最小数据样本要求，则继续切分
        if len(left) >= min_sample and position not in [0.0, 1.0]:
            cutting_data(left, var, target, min_sample, best_bincut)
        else:
            pass
        if len(right) >= min_sample and position not in [0.0, 1.0]:
            cutting_data(right, var, target, min_sample, best_bincut)
        else:
            pass
        return best_bincut

    best_bincut = cutting_data(data, var, target, min_sample, best_bincut)

    # 把切分点补上头尾
    best_bincut.append(data[var].min())
    best_bincut.append(data[var].max())
    best_bincut_set = set(best_bincut)
    best_bincut = list(best_bincut_set)

    best_bincut.remove(data[var].min())
    best_bincut.append(data[var].min() - 1)
    # 排序切分点
    best_bincut.sort()

    return best_bincut





# 等频分箱
def bin_frequency(x, y, n):
    total = y.count()
    bad = y.sum()
    good = total - bad
    d1 = pd.DataFrame({"x": x, "y": y, 'bin': pd.qcut(x, n, duplicates='drop')})
    d2 = d1.groupby('bin', as_index=True)
    d3 = pd.DataFrame()
    d3['total'] = d2.y.count()  ##每个箱中的总样本数
    d3['bad'] = d2.y.sum()  ##每个箱中的坏样本数
    d3['good'] = d3['total'] - d3['bad']  ##每个箱中的好样本数
    d3['bad_rate'] = d3['bad'] / d3['total'] * 100  ##每个箱中的坏样本率
    d3['%bad'] = d3['bad'] / bad * 100  ##每个箱中的坏样本占总坏样本的比重
    d3['%good'] = d3['good'] / good * 100  ##每个箱中的好样本占总好样本的比重
    d3['%cum_bad'] = d3['%bad'].cumsum()
    d3['%cum_good'] = d3['%good'].cumsum()
    d3['woe'] = np.log(d3['%bad'] / d3['%good'])
    d3['bin_iv'] = (d3['%bad'] - d3['%good']) * d3['woe']
    d3['iv'] = d3['bin_iv'].sum()
    d3['bin_ks'] = d3['%cum_bad'] - d3['%cum_good']
    d3['ks'] = d3['bin_ks'].max()
    d3.reset_index(inplace=True)
    return d3


'''
    total = df[target].count()
    bad = df[target].sum()
    good = total-bad
    all_odds = good/bad
    inf = float('inf')
    ninf = float('-inf')
    bin_df=[]
    iv_value=[]
    d1 = df.groupby(bucket)
    d2 = pd.DataFrame()
    d2['min_bin'] = d1[col].min()
    d2['max_bin'] = d1[col].max()
    d2['total'] = d1[target].count()
    d2['totalrate'] = d2['total']/total
    d2['bad'] = d1[target].sum()
    d2['badrate'] = d2['bad']/d2['total']
    d2['good'] = d2['total'] - d2['bad']
    d2['goodrate'] = d2['good']/d2['total']
    d2['badattr'] = d2['bad']/bad
    d2['goodattr'] = (d2['total']-d2['bad'])/good
    d2['cumgoodrate'] =(d2['goodrate']*d2['totalrate']/(good/total)).cumsum()
    d2['cumbadrate'] =(d2['badrate']*d2['totalrate']/(bad/total)).cumsum()
    d2['odds'] = d2['good']/d2['bad']
    GB_list=[]
    for i in d2.odds:
        if i>=all_odds:
            GB_index = str(round((i/all_odds)*100,0))+str('G')
        else:
            GB_index = str(round((all_odds/i)*100,0))+str('B')
        GB_list.append(GB_index)
    d2['GB_index'] = GB_list
    d2['woe'] = np.log(d2['badattr']/d2['goodattr'])
    d2['bin_iv'] = (d2['badattr']-d2['goodattr'])*d2['woe']
    d2['IV'] = d2['bin_iv'].sum()
    iv = d2['bin_iv'].sum().round(3)
    print('变量名:{}'.format(col))
    print('IV:{}'.format(iv))
    print('\t')
    bin_df.append(d2)
    iv_value.append(iv)
    d2.reset_index(inplace=True)
'''


# 等距分箱
def bin_distance(x, y, n=10):  ##主要woe有可能为-inf
    total = y.count()
    bad = y.sum()
    good = total - bad
    d1 = pd.DataFrame({"x": x, "y": y, 'bin': pd.cut(x, n)})  ##等距分箱
    d2 = d1.groupby('bin', as_index=True)
    d3 = pd.DataFrame()
    d3['total'] = d2.y.count()  ##每个箱中的总样本数
    d3['bad'] = d2.y.sum()  ##每个箱中的坏样本数
    d3['good'] = d3['total'] - d3['bad']  ##每个箱中的好样本数
    d3['bad_rate'] = d3['bad'] / d3['total'] * 100  ##每个箱中的坏样本率
    d3['%bad'] = d3['bad'] / bad * 100  ##每个箱中的坏样本占总坏样本的比重
    d3['%good'] = d3['good'] / good * 100  ##每个箱中的好样本占总好样本的比重
    d3['%cum_bad'] = d3['%bad'].cumsum()
    d3['%cum_good'] = d3['%good'].cumsum()
    d3['woe'] = np.log(d3['%bad'] / d3['%good'])
    d3['bin_iv'] = (d3['%bad'] - d3['%good']) * d3['woe']
    d3['iv'] = d3['bin_iv'].sum()
    d3['bin_ks'] = d3['%cum_bad'] - d3['%cum_good']
    d3['ks'] = d3['bin_ks'].max()
    d3.reset_index(inplace=True)
    return d3


# 单调分箱
def mono_bin(x, y, n=20):
    r = 0
    while np.abs(r) < 1:
        d1 = pd.DataFrame({'x': x,
                           'y': y,
                           'bin': pd.qcut(x, n)})
        d2 = d1.groupby('bin', as_index=True)
        print(d1)
        r, p = spearmanr(d2.mean().x, d2.mean().y)
        print(r, p)
        n = n + 1
        print(n)
    total = y.count()
    bad = y.sum()
    good = total - bad
    d3 = pd.DataFrame()
    d3['total'] = d2.y.count()  ##每个箱中的总样本数
    d3['bad'] = d2.y.sum()  ##每个箱中的坏样本数
    d3['good'] = d3['total'] - d3['bad']  ##每个箱中的好样本数
    d3['bad_rate'] = d3['bad'] / d3['total'] * 100  ##每个箱中的坏样本率
    d3['%bad'] = d3['bad'] / bad * 100  ##每个箱中的坏样本占总坏样本的比重
    d3['%good'] = d3['good'] / good * 100  ##每个箱中的好样本占总好样本的比重
    d3['%cum_bad'] = d3['%bad'].cumsum()
    d3['%cum_good'] = d3['%good'].cumsum()
    d3['woe'] = np.log(d3['%bad'] / d3['%good'])
    d3['bin_iv'] = (d3['%bad'] - d3['%good']) * d3['woe']
    d3['iv'] = d3['bin_iv'].sum()
    d3['bin_ks'] = d3['%cum_bad'] - d3['%cum_good']
    d3['ks'] = d3['bin_ks'].max()
    d3.reset_index(inplace=True)
    return d3


# 自定义分箱
def bin_self(x, y, cut):  ##cut:自定义分箱（list）
    total = y.count()
    bad = y.sum()
    good = total - bad
    d1 = pd.DataFrame({"x": x, "y": y, 'bin': pd.cut(x, cut)})  ##等距分箱
    d2 = d1.groupby('bin', as_index=True)
    d3 = pd.DataFrame()
    d3['total'] = d2.y.count()  ##每个箱中的总样本数
    d3['bad'] = d2.y.sum()  ##每个箱中的坏样本数
    d3['good'] = d3['total'] - d3['bad']  ##每个箱中的好样本数
    d3['bad_rate'] = d3['bad'] / d3['total'] * 100  ##每个箱中的坏样本率
    d3['%bad'] = d3['bad'] / bad * 100  ##每个箱中的坏样本占总坏样本的比重
    d3['%good'] = d3['good'] / good * 100  ##每个箱中的好样本占总好样本的比重
    d3['%cum_bad'] = d3['%bad'].cumsum()
    d3['%cum_good'] = d3['%good'].cumsum()
    d3['woe'] = np.log(d3['%bad'] / d3['%good'])
    d3['bin_iv'] = (d3['%bad'] - d3['%good']) * d3['woe']
    d3['iv'] = d3['bin_iv'].sum()
    d3['bin_ks'] = d3['%cum_bad'] - d3['%cum_good']
    d3['ks'] = d3['bin_ks'].max()
    d3.reset_index(inplace=True)
    return d3



# Examples:

# 先卡方分箱
# a = get_chimerge_bincut(df, '决策分数', 'target')
# 再自定义画分箱
# bins_chimerge = bin_self(df['决策分数'], df['target'], a)

# a = get_bestks_bincut(df, '决策分数', 'target')
# bins_bestks = bin_self(df['决策分数'], df['target'], a)





# 测试
'''
df['score_bins1'] = pd.cut(df['决策分数'], bins=get_cart_bincut(df, '决策分数', 'target'))
df['score_bins2'] = pd.cut(df['决策分数'], bins=get_chimerge_bincut(df, '决策分数', 'target'))
df['score_bins3'] = pd.cut(df['决策分数'], bins=get_bestks_bincut(df, '决策分数', 'target'))
print("变量 决策分数 的分箱结果如下：")
print("score_cart_bins:", get_cart_bincut(df, 'score', 'target'))
print("score_chimerge_bins:", get_chimerge_bincut(df, '决策分数', 'target'))
print("score_bestks_bins:", get_bestks_bincut(df, 'score', 'target'))
print("IV值如下：")
print("score:", iv_count(df, '决策分数', 'target'))
print("score_cart_bins:", iv_count(df, 'score_bins1', 'target'))
print("score_chimerge_bins:", iv_count(df, 'score_bins2', 'target'))
print("score_bestks_bins:", iv_count(df, 'score_bins3', 'target'))


df['income_bins1'] = pd.cut(df['income'], bins=get_cart_bincut(df, 'income', 'target'))
df['income_bins2'] = pd.cut(df['income'], bins=get_chimerge_bincut(df, 'income', 'target'))
df['income_bins3'] = pd.cut(df['income'], bins=get_bestks_bincut(df, 'income', 'target'))
print("变量 income 的分箱结果如下：")
print("income_cart_bins:", get_cart_bincut(df, 'income', 'target'))
print("income_chimerge_bins:", get_chimerge_bincut(df, 'income', 'target'))
print("income_bestks_bins:", get_bestks_bincut(df, 'income', 'target'))
print("IV值如下：")
print("income:", iv_count(df, 'income', 'target'))
print("income_cart_bins:", iv_count(df, 'income_bins1', 'target'))
print("income_chimerge_bins:", iv_count(df, 'income_bins2', 'target'))
print("income_bestks_bins:", iv_count(df, 'income_bins3', 'target'))
'''




# 在 main.py 或者 __init__.py 中创建主类 Yihui
import warnings
import pandas as pd

from yihui import Yihui



# 在主程序中使用 Yihui 类
if __name__ == "__main__":
    # ban FutureWarning
    warnings.filterwarnings('ignore')

    # generate data.csv
    with open("../Data/data.csv", "r") as f:
        data = pd.read_csv(f)
    data['customer_no'] = str(data['customer_no'])

    # loading Titanic dataset
    # data = sns.load_dataset('titanic')
    # data.head()

    # Create Yihui Class
    yihui_project = Yihui(data,'dlq_flag')
    print(yihui_project.data.head())

    print("Categorical Variables:", yihui_project.categorical_vars)
    print("Numeric Variables:", yihui_project.numeric_vars)

    # 直接访问 Yihui 类的属性获取字符型和数值型变量的名字
    categorical_vars_list = yihui_project.categorical_vars
    numeric_vars_list = yihui_project.numeric_vars

    # yihui_project.eda_module.auto_eda_profiling()

    # yihui_project.eda_module.plot_num_col(numeric_vars_list,plt_type='hist',hspace=0.4,wspace=0.4,plt_size=(100,100),plt_num=100,x=10,y=10)
    # yihui_project.eda_module.plot_num_col(numeric_vars_list,plt_type='box',hspace=0.4,wspace=0.4,plt_size=(100,100),plt_num=100,x=10,y=10)



    # eda_module.plot_default_num
    # yihui_project.eda_module.plot_default_num(numeric_vars_list,hspace=0.4,wspace=0.4,q=10,plt_size=(100,100),plt_num=100,x=10,y=10)

    # eda_module.plot_default_cate
    # yihui_project.eda_module.plot_default_cate(categorical_vars_list,hspace=0.4,wspace=0.4,plt_size=(10,10),plt_num=1,x=1,y=1)



import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from ydata_profiling import ProfileReport

import pandas as pd
import numpy as np
from scipy.stats import entropy

class EDAModule:
    def __init__(self, yihui_instance):
        self.yihui_instance = yihui_instance
        self.data = self.yihui_instance.data.copy()
        self.variables = self.yihui_instance.data.columns
        self.category_variables = self.yihui_instance.get_categorical_variables()
        self.numeric_variables = self.yihui_instance.get_numeric_variables()

    # 类别型变量的分布
    def plot_cate_var(self, col_list, hspace=0.4, wspace=0.4, plt_size=None, plt_num=None, x=None, y=None):
        """
        self.yihui_instance.data:数据集
        col_list:变量list集合
        hspace :子图之间的间隔(y轴方向)
        wspace :子图之间的间隔(x轴方向)
        plt_size :图纸的尺寸
        plt_num :子图的数量
        x :子图矩阵中一行子图的数量
        y :子图矩阵中一列子图的数量

        return :变量的分布图（柱状图形式）
        """
        plt.figure(figsize=plt_size)
        plt.subplots_adjust(hspace=hspace, wspace=wspace)
        plt.rcParams['font.sans-serif'] = ['Microsoft YaHei']
        plt.rcParams['axes.unicode_minus'] = False
        for i, col in zip(range(1, plt_num + 1, 1), col_list):
            plt.subplot(x, y, i)
            plt.title(col)
            sns.countplot(data=self.yihui_instance.data, y=col)
            plt.ylabel('')
        return plt.show()

    # 数值型变量的分布
    def plot_num_col(self, col_list, plt_type='hist', hspace=0.4, wspace=0.4, plt_size=None, plt_num=None, x=None,
                     y=None):
        """
        col_list:变量list集合
        hspace :子图之间的间隔(y轴方向)
        wspace :子图之间的间隔(x轴方向)
        plt_type: 选择直方图/箱线图
        plt_size :图纸的尺寸
        plt_num :子图的数量
        x :子图矩阵中一行子图的数量
        y :子图矩阵中一列子图的数量

        return :变量的分布图（箱线图/直方图）
        """
        plt.figure(figsize=plt_size)
        plt.subplots_adjust(hspace=hspace, wspace=wspace)
        if plt_type == 'hist':
            for i, col in zip(range(1, plt_num + 1, 1), col_list):
                plt.subplot(x, y, i)
                plt.title(col)
                sns.distplot(self.yihui_instance.data[col].dropna())
                plt.xlabel('')
        if plt_type == 'box':
            for i, col in zip(range(1, plt_num + 1, 1), col_list):
                plt.subplot(x, y, i)
                plt.title(col)
                sns.boxplot(data=self.yihui_instance.data, x=col, fliersize=5,
                            flierprops={'markerfacecolor': 'cornflowerblue', 'markeredgecolor': 'cornflowerblue',
                                        'markersize': 4}, )
                plt.xlabel('')
        if plt_type == 'stripplot':
            for i, col in zip(range(1, plt_num + 1, 1), col_list):
                plt.subplot(x, y, i)
                plt.title(col)
                sns.stripplot(data=self.yihui_instance.data, x=col)
                plt.xlabel('')
        return plt.show()

    # 类别型变量的违约率分析
    def plot_default_cate(self, col_list, hspace=0.4, wspace=0.4, plt_size=None, plt_num=None, x=None, y=None):
        """
        col_list:变量list集合
        hspace :子图之间的间隔(y轴方向)
        wspace :子图之间的间隔(x轴方向)
        plt_size :图纸的尺寸
        plt_num :子图的数量
        x :子图矩阵中一行子图的数量
        y :子图矩阵中一列子图的数量

        return :违约率分布图（柱状图形式）
        """

        all_bad = self.yihui_instance.data[self.yihui_instance.target].sum()
        total = self.yihui_instance.data[self.yihui_instance.target].count()
        all_default_rate = all_bad * 1.0 / total

        plt.figure(figsize=plt_size)
        plt.subplots_adjust(hspace=hspace, wspace=wspace)
        plt.rcParams['font.sans-serif'] = ['Microsoft YaHei']
        plt.rcParams['axes.unicode_minus'] = False
        for i, col in zip(range(1, plt_num + 1, 1), col_list):
            d1 = self.yihui_instance.data.groupby(col)
            d2 = pd.DataFrame()
            d2['total'] = d1[self.yihui_instance.target].count()
            d2['bad'] = d1[self.yihui_instance.target].sum()
            d2['default_rate'] = d2['bad'] / d2['total']
            d2 = d2.reset_index()
            plt.subplot(x, y, i)
            plt.title(col)
            plt.axvline(x=all_default_rate)
            sns.barplot(data=d2, y=col, x='default_rate')
            plt.ylabel('')
        return plt.show()

    # 数值型变量的违约率分析
    def plot_default_num(self, col_list, hspace=0.4, wspace=0.4, q=None, plt_size=None, plt_num=None, x=None,
                         y=None):
        """
        self.yihui_instance.data:数据集
        col_list:变量list集合
        self.yihui_instance.target ：目标变量的字段名
        hspace :子图之间的间隔(y轴方向)
        wspace :子图之间的间隔(x轴方向)
        q :等深分箱的箱体个数
        plt_size :图纸的尺寸
        plt_num :子图的数量
        x :子图矩阵中一行子图的数量
        y :子图矩阵中一列子图的数量

        return :违约率分布图（折线图形式）
        """
        all_bad = self.yihui_instance.data[self.yihui_instance.target].sum()
        total = self.yihui_instance.data[self.yihui_instance.target].count()
        all_default_rate = all_bad * 1.0 / total

        plt.figure(figsize=plt_size)
        plt.subplots_adjust(hspace=hspace, wspace=wspace)
        for i, col in zip(range(1, plt_num + 1, 1), col_list):
            bucket = pd.qcut(self.yihui_instance.data[col], q=q, duplicates='drop')
            d1 = self.yihui_instance.data.groupby(bucket)
            d2 = pd.DataFrame()
            d2['total'] = d1[self.yihui_instance.target].count()
            d2['bad'] = d1[self.yihui_instance.target].sum()
            d2['default_rate'] = d2['bad'] / d2['total']
            d2 = d2.reset_index()
            plt.subplot(x, y, i)
            plt.title(col, color='cornflowerblue')
            plt.axhline(y=all_default_rate)
            sns.pointplot(data=d2, x=col, y='default_rate', color='cornflowerblue')
            plt.xticks(rotation=60)
            plt.xlabel('')
        return plt.show()

    # 使用ydata_profiling进行自动EDA，维度较高时，速度较慢
    def auto_eda_profiling(self):
        # 使用pandas profiling进行自动EDA
        profile = ProfileReport(self.yihui_instance.yihui_instance.data,
                                title="Report",
                                correlations={"auto": {"calculate": False}},
                                missing_diagrams={"Heatmap": False}
                                )  # object created
        profile.to_file(output_file='../Data/output.html')



    def __calculate_category_stats(self):
        category_stats = {}
        for var in self.category_variables:
            unique_count = self.data[var].nunique()
            entropy_val = entropy(self.data[var].value_counts(normalize=True))
            missing_pct = self.data[var].isnull().mean() * 100

            category_stats[var] = {
                'unique_count': unique_count,
                'entropy': entropy_val,
                'missing_pct': missing_pct
            }

        return pd.DataFrame(category_stats).transpose()

    def __calculate_numeric_stats(self):
        numeric_stats = {}
        for var in self.numeric_variables:
            mean_val = self.data[var].mean()
            min_val = self.data[var].min()
            q1 = self.data[var].quantile(0.25)
            median_val = self.data[var].median()
            q3 = self.data[var].quantile(0.75)
            max_val = self.data[var].max()
            missing_pct = self.data[var].isnull().mean() * 100

            numeric_stats[var] = {
                'mean': mean_val,
                'min': min_val,
                'q1': q1,
                'median': median_val,
                'q3': q3,
                'max': max_val,
                'missing_pct': missing_pct
            }

        return pd.DataFrame(numeric_stats).transpose()

    # 快速自动分析数据集（无图）
    def auto_eda_simple(self):
        category_stats = self.__calculate_category_stats()
        numeric_stats = self.__calculate_numeric_stats()

        # Combine the results into a single DataFrame
        eda_results = pd.concat([category_stats, numeric_stats])

        return eda_results



from Yihui.binning import BinningModule
from Yihui.eda import EDAModule
from Yihui.data_processing import DataProcessingModule
from Yihui.cluster import ClusterMuodule
from Yihui.var_selelct import VarSelectModule

import numpy as np
import pandas as pd

# 导入相关库


class Yihui:
    def __init__(self, data, target=None):
        self.data = data
        self.target = target
        self.eda_module = EDAModule(self)
        self.dp_module = DataProcessingModule(self)
        self.cluster_module = ClusterMuodule(self)
        self.binning_module = BinningModule(self)
        self.var_select_module = VarSelectModule(self)
        # 其他模块的初始化...

    # 提取字符型变量的名字并返回一个list
    def get_categorical_variables(self):
        # cate_vars = list(self.data.select_dtypes(include='object').columns)
        cate_vars = self.data.select_dtypes(include=[np.object, 'category']).columns.tolist()
        if self.data[self.target].dtype == 'object':
            cate_vars.remove(self.target)
        return cate_vars

    # 提取数值型变量的名字并返回一个list
    def get_numeric_variables(self):
        # num_vars = list(self.data.select_dtypes(exclude='object').columns)
        num_vars = self.data.select_dtypes(include=[np.number]).columns.tolist()

        # 剔除日期型变量
        date_vars = self.get_date_variables()
        numeric_vars = [var for var in num_vars if var not in date_vars]

        if self.data[self.target].dtype == np.number:
            num_vars.remove(self.target)
        return num_vars

    def get_date_variables(self):
        date_vars = []

        for col in self.data.columns:
            if self.is_numeric_date_format(col):
                date_vars.append(col)

        return date_vars

    def is_numeric_date_format(self, col):
        if self.data[col].dropna().empty:
            return False

        date_formats = ['%Y%m%d', '%Y%m', '%m%d', '%Y%m%d%H%M%S', '%Y-%m-%d', '%Y-%m', '%m-%d']

        for date_format in date_formats:
            try:
                # 尝试转换为日期
                pd.to_datetime(self.data[col], format=date_format, errors='raise')
                return True
            except (ValueError, TypeError):
                continue

        return False



from setuptools import setup
from setuptools import find_packages

VERSION = '0.1.10'

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='Yihuier',  # package name
    version=VERSION,  # package version
    author='encyc',
    author_email='atomyuangao@gmail.com',
    description='Package for Logistic Regression Modeling, focus on Credit Risk Management',  # package description
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/encyc/yihui',
    include_package_data=True,
    packages=find_packages(),
    zip_safe=False,
    python_requires='>=3.6',

    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Build Tools",
        "Intended Audience :: Developers"
    ],
    install_requires=[
        'pandas',
        'numpy',
        'matplotlib',
        'seaborn',
        'scikit-learn',
        'ydata_profiling==4.6.3',
    ],

    # data_files=[
    #     ('', ['conf/*.conf']),
    #     ('/usr/lib/systemd/system/', ['bin/*.service']),
    # ],
    # exclude_package_data={
    #     'bandwidth_reporter': ['*.txt']
    # },
    package_data={
        'Data': ['*.csv']
    }
)
